#!/usr/bin/env python3
"""
SPDX-License-Identifier: Apache-2.0
Copyright Contributors to the ODPi Egeria project.

Display by deployed implementation type.

Note: This implementation is using the runtime manager.



A simple server status display
"""
import argparse
import os
import sys
import time

from rich.console import Console
from rich.markdown import Markdown
from rich.prompt import Prompt
from rich.table import Table
from rich.text import Text

from pyegeria import (
    PyegeriaAPIException, PyegeriaClientException,
    RuntimeManager,
    print_basic_exception,
    settings,
    config_logging, PyegeriaException
)


EGERIA_USER = os.environ.get("EGERIA_USER", "erinoverview")
EGERIA_USER_PASSWORD = os.environ.get("EGERIA_USER_PASSWORD", "secret")
app_config = settings.Environment
# config_logging()


def display_servers_by_dep_imp(search_string: str = "*", view_server: str = app_config.egeria_view_server,
                               view_url: str = app_config.egeria_view_server_url, user: str = EGERIA_USER,
                               user_pass: str = EGERIA_USER_PASSWORD, jupyter: bool = app_config.egeria_jupyter,
                               width: int = app_config.console_width):
    p_client = RuntimeManager(view_server, view_url, user, user_pass)
    token = p_client.create_egeria_bearer_token()

    def generate_table() -> Table:
        """Make a new table."""
        table = Table(
            title=f"Servers by Deployed Implementation Type - {time.asctime()}",
            style="bold white on black",
            row_styles=["bold white on black"],
            header_style="white on dark_blue",
            title_style="bold white on black",
            caption_style="white on black",
            caption=f"Platform - '{view_url}'",
            show_lines=True,
        )

        table.add_column("Server")
        table.add_column("Deployed Impl Type")
        table.add_column("Zones")

        table.add_column("Description")
        table.add_column("Qualified Name & GUID", no_wrap=True)

        unsorted_server_list = p_client.get_servers_by_dep_impl_type(search_string)
        if type(unsorted_server_list) is str:
            print("No matching Software Servers found?")
            sys.exit(1)
        server_list = sorted(
            unsorted_server_list,
            key=lambda x: x["properties"].get("displayName", "---").lower(),
        )

        for server in server_list:
            display_name = server["properties"].get("displayName", "---")
            qualified_name = server["properties"].get("qualifiedName", "---")
            classifications = server["elementHeader"].get("classifications", None)
            zones = ""
            # if classifications:
            #     for clas in classifications:
            #         if clas["classificationName"] == "AssetZoneMembership":
            #             classification_props = clas["classificationProperties"]
            #             zone_membership = classification_props.get(
            #                 "zoneMembership", None
            #             )
            #             if zone_membership:
            #                 for z in zone_membership.keys():
            #                     zones += f"{zone_membership[z]}, "
            #                 zones = zones[:-2]

            impl_type = server["properties"].get("deployedImplementationType", "---")
            server_guid = server["elementHeader"]["guid"]
            server_desc = server["properties"].get("description", "---")

            server_id = Text(f"{qualified_name}\n&\n{server_guid}", justify="center")

            table.add_row(display_name, impl_type, zones, server_desc, server_id)

        return table

    try:
        console = Console(width=width, force_terminal=not jupyter)

        with console.pager(styles=True):
            console.print(generate_table())

    except (
        PyegeriaException
    ) as e:
        print_basic_exception(e)
    except KeyboardInterrupt:
        pass
    finally:
        p_client.close_session()


def main():
    parser = argparse.ArgumentParser()

    parser.add_argument("--server", help="Name of the view server")
    parser.add_argument("--url", help="URL Platform to connect to")
    parser.add_argument("--userid", help="User Id")
    parser.add_argument("--password", help="User Password")
    args = parser.parse_args()

    server = args.server if args.server is not None else app_config.egeria_view_server
    url = args.url if args.url is not None else app_config.egeria_view_server_url
    userid = args.userid if args.userid is not None else EGERIA_USER
    user_pass = args.password if args.password is not None else EGERIA_USER_PASSWORD
    filter_string = Prompt.ask(
        "Name of deployed server type", default="View Server"
    )
    display_servers_by_dep_imp(filter_string, server, url, userid, user_pass)


if __name__ == "__main__":
    main()



"""
SPDX-License-Identifier: Apache-2.0
Copyright Contributors to the ODPi Egeria project.

This module tests the SecurityOfficer class and methods from security_officer.py
"""

import json
import time
from datetime import datetime

from rich import print, print_json
from rich.console import Console

from pyegeria.core._exceptions import (
    PyegeriaInvalidParameterException,
    PyegeriaConnectionException,
    PyegeriaAPIException,
    PyegeriaNotFoundException,
    PyegeriaUnauthorizedException,
    PyegeriaClientException,
    print_basic_exception,
    print_exception_table,
)
from pyegeria.omvs.security_officer import SecurityOfficer
from pyegeria.core.logging_configuration import config_logging, init_logging
from pyegeria.models import (
    UserAccountRequestBody,
    OpenMetadataUserAccount,
    NewRelationshipRequestBody,
    DeleteRelationshipRequestBody,
)

disable_ssl_warnings = True

console = Console(width=250)

config_logging()
init_logging(True)

VIEW_SERVER = "qs-view-server"
PLATFORM_URL = "https://localhost:9443"
PLATFORM_NAME = "Local OMAG Server Platform"
USER_ID = "freddiemercury"
USER_PWD = "magic, magic"


class TestSecurityOfficer:
    good_platform1_url = PLATFORM_URL
    good_platform_name = PLATFORM_NAME
    good_user_1 = USER_ID
    good_server_1 = VIEW_SERVER

    def test_set_user_account(self):
        """Test creating/updating a user account"""
        try:
            s_client = SecurityOfficer(
                self.good_server_1, self.good_platform1_url, user_id=self.good_user_1
            )
            s_client.create_egeria_bearer_token(self.good_user_1, USER_PWD)

            user_id = "testuser" + datetime.now().strftime("%H%M%S")
            body = {
                "class": "UserAccountRequestBody",
                "userAccount": {
                    "class": "OpenMetadataUserAccount",
                    "userId": user_id,
                    "userName": "Test User",
                    "userAccountStatus": "AVAILABLE",
                    "secrets": {"clearPassword": "testpassword"},
                },
            }

            s_client.set_user_account(self.good_platform_name, body)
            print(f"\n\nSet user account for: {user_id}")
            assert True

        except (
            PyegeriaInvalidParameterException,
            PyegeriaAPIException,
            PyegeriaUnauthorizedException,
            PyegeriaNotFoundException,
        ) as e:
            print_exception_table(e)
            assert False, "Invalid request"
        except PyegeriaConnectionException as e:
            print_basic_exception(e)
            assert False, "Connection error"
        finally:
            s_client.close_session()

    def test_get_user_account(self):
        """Test retrieving a user account"""
        try:
            s_client = SecurityOfficer(
                self.good_server_1, self.good_platform1_url, user_id=self.good_user_1
            )
            s_client.create_egeria_bearer_token(self.good_user_1, USER_PWD)

            # Retrieve Gary Geeke's account as a test (assuming it exists)
            user_id = "garygeeke"
            response = s_client.get_user_account(self.good_platform_name, user_id)
            print(f"\n\nRetrieved user account for: {user_id}")
            print_json(data=response)
            assert type(response) is dict
            assert response["userId"] == user_id

        except (
            PyegeriaInvalidParameterException,
            PyegeriaAPIException,
            PyegeriaUnauthorizedException,
            PyegeriaNotFoundException,
        ) as e:
            print_exception_table(e)
            assert False, "Invalid request"
        except PyegeriaConnectionException as e:
            print_basic_exception(e)
            assert False, "Connection error"
        finally:
            s_client.close_session()

    def test_delete_user_account(self):
        """Test deleting a user account"""
        try:
            s_client = SecurityOfficer(
                self.good_server_1, self.good_platform1_url, user_id=self.good_user_1
            )
            s_client.create_egeria_bearer_token(self.good_user_1, USER_PWD)

            user_id = "deleteuser" + datetime.now().strftime("%H%M%S")
            # Create first
            body = {
                "class": "UserAccountRequestBody",
                "userAccount": {
                    "class": "OpenMetadataUserAccount",
                    "userId": user_id,
                    "userName": "Delete Test User",
                },
            }
            s_client.set_user_account(self.good_platform_name, body)

            # Then delete
            s_client.delete_user_account(self.good_platform_name, user_id)
            print(f"\n\nDeleted user account: {user_id}")
            assert True

        except (
            PyegeriaInvalidParameterException,
            PyegeriaAPIException,
            PyegeriaUnauthorizedException,
            PyegeriaNotFoundException,
        ) as e:
            print_exception_table(e)
            assert False, "Invalid request"
        except PyegeriaConnectionException as e:
            print_basic_exception(e)
            assert False, "Connection error"
        finally:
            s_client.close_session()

    def test_link_governance_zones(self):
        """Test linking governance zones"""
        try:
            s_client = SecurityOfficer(
                self.good_server_1, self.good_platform1_url, user_id=self.good_user_1
            )
            s_client.create_egeria_bearer_token(self.good_user_1, USER_PWD)

            # These GUIDs are placeholders for unit test structure
            zone_guid = "zone-123"
            nested_zone_guid = "zone-456"
            body = {
                "class": "NewRelationshipRequestBody",
                "properties": {
                    "class": "ZoneHierarchy",
                },
            }

            # This will likely fail with 404/NotFound if GUIDs don't exist, but tests structure
            try:
                s_client.link_governance_zones(zone_guid, nested_zone_guid, body)
                assert True
            except (PyegeriaNotFoundException, PyegeriaAPIException, PyegeriaClientException):
                print("Caught expected exception for placeholder GUIDs")
                assert True

        except (
            PyegeriaInvalidParameterException,
            PyegeriaAPIException,
            PyegeriaUnauthorizedException,
        ) as e:
            print_exception_table(e)
            assert False, "Invalid request"
        except PyegeriaConnectionException as e:
            print_basic_exception(e)
            assert False, "Connection error"
        finally:
            s_client.close_session()

    def test_detach_governance_zones(self):
        """Test detaching governance zones"""
        try:
            s_client = SecurityOfficer(
                self.good_server_1, self.good_platform1_url, user_id=self.good_user_1
            )
            s_client.create_egeria_bearer_token(self.good_user_1, USER_PWD)

            zone_guid = "zone-123"
            nested_zone_guid = "zone-456"
            body = {
                "class": "DeleteRelationshipRequestBody",
            }

            try:
                s_client.detach_governance_zones(zone_guid, nested_zone_guid, body)
                assert True
            except (PyegeriaNotFoundException, PyegeriaAPIException, PyegeriaClientException):
                print("Caught expected exception for placeholder GUIDs")
                assert True

        except (
            PyegeriaInvalidParameterException,
            PyegeriaAPIException,
            PyegeriaUnauthorizedException,
        ) as e:
            print_exception_table(e)
            assert False, "Invalid request"
        except PyegeriaConnectionException as e:
            print_basic_exception(e)
            assert False, "Connection error"
        finally:
            s_client.close_session()



from typing import Callable, Dict, Any, Optional
from loguru import logger

class CommandDispatcher:
    """
    Registry and dispatcher for Dr.Egeria markdown commands.
    """
    def __init__(self):
        self._registry: Dict[str, Callable] = {}

    def register(self, command_name: str, handler: Callable):
        """
        Register a handler for a specific command name.
        """
        self._registry[command_name] = handler
        logger.debug(f"Registered command handler for: {command_name}")

    def get_handler(self, command_name: str) -> Optional[Callable]:
        """
        Retrieve the handler for a given command name.
        """
        return self._registry.get(command_name)

    def dispatch(self, command_name: str, *args, **kwargs) -> Any:
        """
        Execute the handler for the given command.
        """
        handler = self.get_handler(command_name)
        if handler:
            return handler(*args, **kwargs)
        else:
            logger.warning(f"No handler registered for command: {command_name}")
            return None



"""
SPDX-License-Identifier: Apache-2.0
Copyright Contributors to the ODPi Egeria project.

Scenario tests for MetadataExpert OMVS.
"""
import time
from datetime import datetime
from rich.console import Console
from pyegeria.omvs.metadata_expert import MetadataExpert
from pyegeria.core._exceptions import PyegeriaException, PyegeriaTimeoutException

VIEW_SERVER = "qs-view-server"
PLATFORM_URL = "https://localhost:9443"
USER_ID = "peterprofile"

console = Console()

def run_scenario():
    console.print("[bold cyan]Starting Metadata Expert Scenario Test[/bold cyan]")
    client = MetadataExpert(VIEW_SERVER, PLATFORM_URL, USER_ID)
    
    try:
        client.create_egeria_bearer_token(USER_ID, "secret")
    except Exception as e:
        console.print(f"[yellow]Warning: Could not create bearer token (server might be down): {e}[/yellow]")

    try:
        # 1. Create Element
        console.print("1. Creating metadata element...")
        qname = f"ScenarioElement-{datetime.now().timestamp()}"
        create_body = {
            "class": "NewOpenMetadataElementRequestBody",
            "typeName": "Referenceable",
            "properties": {
                "class": "ElementProperties",
                "propertyValueMap": {
                    "qualifiedName": {
                        "class": "PrimitiveTypePropertyValue",
                        "typeName": "string",
                        "primitiveValue": qname
                    }
                }
            }
        }
        guid = client.create_metadata_element(create_body)
        console.print(f"   Created element with GUID: [green]{guid}[/green]")
        
        # 2. Update Element
        console.print("2. Updating element properties...")
        update_body = {
            "class": "UpdatePropertiesRequestBody",
            "properties": {
                "class": "ElementProperties",
                "propertyValueMap": {
                    "description": {
                        "class": "PrimitiveTypePropertyValue",
                        "typeName": "string",
                        "primitiveValue": "Scenario description"
                    }
                }
            }
        }
        client.update_metadata_element_properties(guid, update_body)
        console.print("   Updated properties successfully.")
        
        # 3. Classify Element
        console.print("3. Classifying element...")
        classify_body = {
            "class": "NewClassificationRequestBody"
        }
        client.classify_metadata_element(guid, "Confidentiality", classify_body)
        console.print("   Classified element as Confidentiality.")
        
        # 4. Delete Element
        console.print("4. Deleting element...")
        delete_body = {
            "class": "OpenMetadataDeleteRequestBody"
        }
        client.delete_metadata_element(guid, delete_body)
        console.print("   Deleted element successfully.")
        
        console.print("[bold green]Scenario completed successfully![/bold green]")
        
    except PyegeriaTimeoutException as e:
        console.print(f"[bold yellow]⚠ Timeout during scenario; continuing: {e}[/bold yellow]")
    except Exception as e:
        console.print(f"[bold red]Scenario failed: {e}[/bold red]")

def test_metadata_expert_scenarios():
    run_scenario()

if __name__ == "__main__":
    test_metadata_expert_scenarios()



"""
SPDX-License-Identifier: Apache-2.0
Copyright Contributors to the ODPi Egeria project.



This module is for testing the core OMAG config class and methods
The routines assume that pytest is being used as the test tool and framework.

A running Egeria environment is needed to run these tests.

"""

import json
import time

from rich import print, print_json
from rich.console import Console
from pyegeria import (
    ValidMetadataManager,
    print_basic_exception,
    PyegeriaException,
)

# from pyegeria.admin_services import FullServerConfig

disable_ssl_warnings = True
console = Console(width = 300)

class TestValidMetadataOMVs:
    good_platform1_url = "https://127.0.0.1:9443"
    # good_platform2_url = "https://oak.local:9443"
    # bad_platform1_url = "https://localhost:9443"



    good_user_1 = "garygeeke"
    good_user_2 = "erinoverview"
    good_user_3 = "peterprofile"
    bad_user_1 = "eviledna"
    bad_user_2 = ""
    good_integ_1 = "fluffy_integration"
    good_server_1 = "simple-metadata-store"
    good_server_2 = "qs-view-server"
    good_server_3 = "active-metadata-store"
    good_server_4 = "integration-daemon"
    good_server_5 = "fluffy_kv"
    good_server_6 = "cocoVIew1"
    good_engine_host_1 = "governDL01"
    good_view_server_1 = "view-server"
    good_view_server_2 = "qs-view-server"
    bad_server_1 = "coco"
    bad_server_2 = ""

    def test_setup_valid_metadata_value(self):
        try:
            m_client = ValidMetadataManager(
                self.good_view_server_2,
                self.good_platform1_url,
                user_id=self.good_user_2,
            )
            token = m_client.create_egeria_bearer_token(self.good_user_2, "secret")
            start_time = time.perf_counter()
            type_name = "Project"
            property_name = "projectHealth"
            body = {
                "displayName": "Abandoned",
                "description": "Who cares?",
                "preferredValue": "Abandoned",
                "dataType": "string",
                "isCaseSensitive": False,
                "isDeprecated": False,
                "additionalProperties": {"colour": "black"},
            }

            response = m_client.setup_valid_metadata_value(
                property_name, type_name, body
            )
            duration = time.perf_counter() - start_time

            print(f"\n\tDuration was {duration} seconds")
            if isinstance(response, (list, dict)):
                print_json("\n\n" + json.dumps(response, indent=4))
            elif type(response) is tuple:
                print(f"Type is {type(response)}")
                print_json("\n\n" + json.dumps(response, indent=4))
            elif type(response) is str:
                print("\n\nGUID is: " + response)
            assert True

        except PyegeriaException as e:
            print_basic_exception(e)
            assert False, "Invalid request"

        finally:
            m_client.close_session()

    def test_setup_valid_metadata_map_name(self):
        try:
            m_client = ValidMetadataManager(
                self.good_view_server_2,
                self.good_platform1_url,
                user_id=self.good_user_2,
            )
            token = m_client.create_egeria_bearer_token(self.good_user_2, "secret")
            start_time = time.perf_counter()
            type_name = "Project"
            property_name = "additionalProperties"
            body = {
                "displayName": "Expected Duration",
                "description": "How long is this project expected to take?",
                "preferredValue": "expectedDuration",
                "dataType": "string",
                "isCaseSensitive": False,

            }

            response = m_client.setup_valid_metadata_map_name(
                property_name, type_name, body
            )
            duration = time.perf_counter() - start_time

            print(f"\n\tDuration was {duration} seconds")
            if isinstance(response, (list, dict)):
                print_json("\n\n" + json.dumps(response, indent=4))
            elif type(response) is tuple:
                print(f"Type is {type(response)}")
                print_json("\n\n" + json.dumps(response, indent=4))
            elif type(response) is str:
                print("\n\nGUID is: " + response)
            assert True

        except PyegeriaException as e:
            print_basic_exception(e)
            assert False, "Invalid request"

        finally:
            m_client.close_session()

    def test_setup_valid_metadata_map_value(self):
        try:
            m_client = ValidMetadataManager(
                self.good_view_server_2,
                self.good_platform1_url,
                user_id=self.good_user_2,
            )
            token = m_client.create_egeria_bearer_token(self.good_user_2, "secret")
            start_time = time.perf_counter()
            type_name = "Project"
            property_name = "additionalProperties"
            map_name = "expectedDuration"
            body = {
                "displayName": "Two Months",
                "description": "The project is expted to last 2 months.",
                "preferredValue": "Two Months",
                "dataType": "string",
                "isCaseSensitive": False,

            }

            response = m_client.setup_valid_metadata_map_value(
                property_name, map_name, type_name, body
            )
            duration = time.perf_counter() - start_time

            print(f"\n\tDuration was {duration} seconds")
            if isinstance(response, (list, dict)):
                print_json("\n\n" + json.dumps(response, indent=4))
            elif type(response) is tuple:
                print(f"Type is {type(response)}")
                print_json("\n\n" + json.dumps(response, indent=4))
            elif type(response) is str:
                print("\n\nGUID is: " + response)
            assert True

        except PyegeriaException as e:
            print_basic_exception(e)
            assert False, "Invalid request"

        finally:
            m_client.close_session()

    def test_get_valid_metadata_value(self):
        try:
            m_client = ValidMetadataManager(
                self.good_view_server_2,
                self.good_platform1_url,
                user_id=self.good_user_2,
            )
            token = m_client.create_egeria_bearer_token(self.good_user_2, "secret")
            start_time = time.perf_counter()
            type_name = None
            property_name = "fileType"
            preferred_value = None

            response = m_client.get_valid_metadata_value(
                property_name, type_name, preferred_value, output_format="LIST", report_spec='Valid-Values' )
            duration = time.perf_counter() - start_time

            print(f"\n\tDuration was {duration} seconds")
            if isinstance(response, (list, dict)):
                print_json("\n\n" + json.dumps(response, indent=4))
            elif type(response) is tuple:
                print(f"Type is {type(response)}")
                print_json("\n\n" + json.dumps(response, indent=4))
            elif type(response) is str:
                print("\n\nGUID is: " + response)
            assert True

        except PyegeriaException as e:
            print_basic_exception(e)
            assert False, "Invalid request"

        finally:
            m_client.close_session()

    def test_get_valid_metadata_values(self):
        try:
            m_client = ValidMetadataManager(
                self.good_view_server_2,
                self.good_platform1_url,
                user_id=self.good_user_2,
            )
            token = m_client.create_egeria_bearer_token(self.good_user_2, "secret")
            start_time = time.perf_counter()
            type_name = None
            property_name = "fileType"
            # type_name = None
            # property_name = "stewardTypeName"

            response = m_client.get_valid_metadata_values(property_name, type_name, output_format="JSON", report_spec='Valid-Values')
            duration = time.perf_counter() - start_time

            print(f"\n\tDuration was {duration} seconds")
            if isinstance(response, list | dict):
                print_json("\n\n" + json.dumps(response, indent=4))
            elif type(response) is tuple:
                print(f"Type is {type(response)}")
                print_json("\n\n" + json.dumps(response, indent=4))
            elif type(response) is str:
                print("\n\nOutput is: " + response)
            assert True

        except PyegeriaException as e:
            print_basic_exception(e)
            assert False, "Invalid request"

        finally:
            m_client.close_session()

    def test_get_valid_metadata_map_values(self):
        try:
            m_client = ValidMetadataManager(
                self.good_view_server_2,
                self.good_platform1_url,
                user_id=self.good_user_2,
            )
            token = m_client.create_egeria_bearer_token(self.good_user_2, "secret")
            start_time = time.perf_counter()
            type_name = "Project"
            property_name = "additionalProperties"
            preferred_value = "1 month"
            map_name = "expectedDuration"
            # type_name = None
            # property_name = "stewardTypeName"

            response = m_client.get_valid_metadata_map_value(property_name, type_name, preferred_value, map_name,
                                                             output_format="DICT", report_spec='Valid-Values')
            duration = time.perf_counter() - start_time

            print(f"\n\tDuration was {duration} seconds")
            if isinstance(response, list | dict):
                print_json("\n\n" + json.dumps(response, indent=4))
            elif type(response) is tuple:
                print(f"Type is {type(response)}")
                print_json("\n\n" + json.dumps(response, indent=4))
            elif type(response) is str:
                print("\n\nOutput is: " + response)
            assert True

        except PyegeriaException as e:
            print_basic_exception(e)
            assert False, "Invalid request"

        finally:
            m_client.close_session()


    def test_get_all_valid_metadata_values(self):
        try:
            m_client = ValidMetadataManager(
                self.good_view_server_2,
                self.good_platform1_url,
                user_id=self.good_user_2,
            )
            token = m_client.create_egeria_bearer_token(self.good_user_2, "secret")
            start_time = time.perf_counter()
            type_name = "Project"
            property_name = "purposes"
            preferred_value = "1 month"
            map_name = "expectedDuration"
            # type_name = None
            # property_name = "stewardTypeName"

            response = m_client.get_valid_metadata_values(property_name, type_name,
                                                             output_format="DICT", report_spec='Valid-Values')
            duration = time.perf_counter() - start_time

            print(f"\n\tDuration was {duration} seconds")
            if isinstance(response, list | dict):
                print_json("\n\n" + json.dumps(response, indent=4))
            elif type(response) is tuple:
                print(f"Type is {type(response)}")
                print_json("\n\n" + json.dumps(response, indent=4))
            elif type(response) is str:
                console.print("\n\nOutput is: " + response)
            assert True

        except PyegeriaException as e:
            print_basic_exception(e)
            assert False, "Invalid request"

        finally:
            m_client.close_session()


    def test_validate_metadata_value(self):
        try:
            m_client = ValidMetadataManager(
                self.good_view_server_2,
                self.good_platform1_url,
                user_id=self.good_user_2,
            )
            token = m_client.create_egeria_bearer_token(self.good_user_2, "secret")
            start_time = time.perf_counter()
            type_name = "Project"
            property_name = "purposes"
            actual_value = "marketing"


            response = m_client.validate_metadata_value(property_name, type_name, actual_value = actual_value)
            duration = time.perf_counter() - start_time

            print(f"\n\tDuration was {duration} seconds")
            if isinstance(response, list | dict):
                print_json("\n\n" + json.dumps(response, indent=4))
            elif type(response) is tuple:
                print(f"Type is {type(response)}")
                print_json("\n\n" + json.dumps(response, indent=4))
            elif type(response) is str:
                print("\n\nOutput is: " + response)
            else:
                print(response)
            assert True

        except PyegeriaException as e:
            print_basic_exception(e)
            assert False, "Invalid request"

        finally:
            m_client.close_session()

    def test_set_consistent_metadata_values(self):
        try:
            m_client = ValidMetadataManager(
                self.good_view_server_2,
                self.good_platform1_url,
                user_id=self.good_user_2,
            )
            token = m_client.create_egeria_bearer_token(self.good_user_2, "secret")
            start_time = time.perf_counter()

            property_name1 = "category"
            property_name2 = "purposes"
            type_name1 = None
            type_name2 = "Project"
            map_name1 = None
            map_name2 = None
            preferred_value1 = "clinical-trial"
            preferred_value2 = "product-verification"


            response = m_client.set_consistent_metadata_values(property_name1, type_name1, map_name1, preferred_value1,
                                                               property_name2, type_name2, map_name2, preferred_value2)
            duration = time.perf_counter() - start_time

            print(f"\n\tDuration was {duration} seconds")
            if isinstance(response, list | dict):
                print_json("\n\n" + json.dumps(response, indent=4))
            elif type(response) is tuple:
                print(f"Type is {type(response)}")
                print_json("\n\n" + json.dumps(response, indent=4))
            elif type(response) is str:
                print("\n\nOutput is: " + response)
            else:
                print(response)
            assert True

        except PyegeriaException as e:
            print_basic_exception(e)
            assert False, "Invalid request"

        finally:
            m_client.close_session()

    def test_get_consistent_metadata_values(self):
        try:
            m_client = ValidMetadataManager(
                self.good_view_server_2,
                self.good_platform1_url,
                user_id=self.good_user_2,
            )
            token = m_client.create_egeria_bearer_token(self.good_user_2, "secret")
            start_time = time.perf_counter()

            # property_name = "category"
            # type_name = "Project"
            # map_name = None
            # preferred_value = "clinical-trial"
            property_name = 'fileType'
            type_name = 'CSVFile'
            map_name = None
            preferred_value = None
            response = m_client.get_consistent_metadata_values(property_name='fileType',
                                                           type_name='CSVFile',
                                                           map_name=None,
                                                           preferred_value="CSV File")

            # response = m_client.get_consistent_metadata_values(property_name, type_name, map_name, preferred_value)
            duration = time.perf_counter() - start_time

            print(f"\n\tDuration was {duration} seconds")
            if isinstance(response, list | dict):
                print_json("\n\n" + json.dumps(response, indent=4))
            elif type(response) is tuple:
                print(f"Type is {type(response)}")
                print_json("\n\n" + json.dumps(response, indent=4))
            elif type(response) is str:
                print("\n\nOutput is: " + response)
            else:
                print(response)
            assert True

        except PyegeriaException as e:
            print_basic_exception(e)
            assert False, "Invalid request"

        finally:
            m_client.close_session()


    def test_find_spec_property(self):
        try:
            m_client = ValidMetadataManager(
                self.good_view_server_2,
                self.good_platform1_url,
                user_id=self.good_user_2,
            )
            token = m_client.create_egeria_bearer_token(self.good_user_2, "secret")
            start_time = time.perf_counter()
            search_string = "annotation"

            response = m_client.find_specification_property(search_string)
            duration = time.perf_counter() - start_time

            print(f"\n\tDuration was {duration} seconds for {len(response)} elements")
            if isinstance(response, list | dict):
                print_json("\n\n" + json.dumps(response, indent=4))
            elif type(response) is tuple:
                print(f"Type is {type(response)}")
                print_json("\n\n" + json.dumps(response, indent=4))
            elif type(response) is str:
                print("\n\nOutput is: " + response)
            else:
                print(response)
            assert True

        except PyegeriaException as e:
            print_basic_exception(e)
            assert False, "Invalid request"

        finally:
            m_client.close_session()

## Entities, TypeDefs etc.

    def test_get_all_entity_types(self):
        try:
            m_client = ValidMetadataManager(
                self.good_view_server_2,
                self.good_platform1_url,
                user_id=self.good_user_2,
            )
            token = m_client.create_egeria_bearer_token(self.good_user_2, "secret")
            start_time = time.perf_counter()

            response = m_client.get_all_entity_types(output_format="JSON", report_spec="Referenceable")
            duration = time.perf_counter() - start_time

            print(f"\n\tDuration was {duration} seconds")
            if isinstance(response, (list, dict)):
                print_json("\n\n" + json.dumps(response, indent=4))
            elif type(response) is tuple:
                print(f"Type is {type(response)}")
                print_json("\n\n" + json.dumps(response, indent=4))
            elif type(response) is str:
                print("\n\nGUID is: " + response)
            assert True

        except (
            PyegeriaException
        ) as e:
            print_basic_exception(e)
            assert False, "Invalid request"

        finally:
            m_client.close_session()

    def test_get_all_entity_defs(self):
        try:
            m_client = ValidMetadataManager(
                self.good_view_server_2,
                self.good_platform1_url,
                user_id=self.good_user_2,
            )
            token = m_client.create_egeria_bearer_token(self.good_user_2, "secret")
            start_time = time.perf_counter()

            response = m_client.get_all_entity_defs()
            duration = time.perf_counter() - start_time

            print(f"\n\tDuration was {duration} seconds")
            if isinstance(response, (list, dict)):
                print_json("\n\n" + json.dumps(response, indent=4))
            elif type(response) is tuple:
                print(f"Type is {type(response)}")
                print_json("\n\n" + json.dumps(response, indent=4))
            elif type(response) is str:
                print("\n\nGUID is: " + response)
            assert True

        except PyegeriaException as e:
            print_basic_exception(e)
            assert False, "Invalid request"

        finally:
            m_client.close_session()

    def test_get_all_relationship_defs(self):
        try:
            m_client = ValidMetadataManager(
                self.good_view_server_2,
                self.good_platform1_url,
                user_id=self.good_user_2,
            )
            token = m_client.create_egeria_bearer_token(self.good_user_2, "secret")
            start_time = time.perf_counter()

            response = m_client.get_all_relationship_defs()
            duration = time.perf_counter() - start_time

            print(f"\n\tDuration was {duration} seconds")
            if isinstance(response, (list, dict)):
                print_json("\n\n" + json.dumps(response, indent=4))
            elif type(response) is tuple:
                print(f"Type is {type(response)}")
                print_json("\n\n" + json.dumps(response, indent=4))
            elif type(response) is str:
                print("\n\nGUID is: " + response)
            assert True

        except PyegeriaException as e:
            print_basic_exception(e)
            assert False, "Invalid request"

        finally:
            m_client.close_session()

    def test_get_all_classification_defs(self):
        try:
            m_client = ValidMetadataManager(
                self.good_view_server_2,
                self.good_platform1_url,
                user_id=self.good_user_2,
            )
            token = m_client.create_egeria_bearer_token(self.good_user_2, "secret")
            start_time = time.perf_counter()

            response = m_client.get_all_classification_defs(output_format="LIST", report_spec="TypeDef")
            duration = time.perf_counter() - start_time
            duration = time.perf_counter() - start_time

            print(f"\n\tDuration was {duration} seconds")
            if isinstance(response, (list, dict)):
                print_json("\n\n" + json.dumps(response, indent=4))
            elif type(response) is tuple:
                print(f"Type is {type(response)}")
                print_json("\n\n" + json.dumps(response, indent=4))
            elif type(response) is str:
                console.print("\n\nGUID is: " + response)
            assert True

        except PyegeriaException as e:
            print_basic_exception(e)
            assert False, "Invalid request"

        finally:
            m_client.close_session()

    def test_get_sub_types(self):
        try:
            m_client = ValidMetadataManager(
                self.good_view_server_2,
                self.good_platform1_url,
                user_id=self.good_user_2,
            )
            token = m_client.create_egeria_bearer_token(self.good_user_2, "secret")
            start_time = time.perf_counter()
            type_name = "Collection"
            response = m_client.get_sub_types(type_name, output_format="MERMAID",report_spec="Common-Mermaid")
            duration = time.perf_counter() - start_time

            print(f"\n\tDuration was {duration} seconds")
            if isinstance(response, (list, dict)):
                print_json("\n\n" + json.dumps(response, indent=4))
            elif type(response) is tuple:
                print(f"Type is {type(response)}")
                print_json("\n\n" + json.dumps(response, indent=4))
            elif type(response) is str:
                print("\n\nGUID is: " + response)
            assert True

        except PyegeriaException as e:
            print_basic_exception(e)
            assert False, "Invalid request"

        finally:
            m_client.close_session()




    def test_get_valid_relationship_types(self):
        try:
            m_client = ValidMetadataManager(
                self.good_view_server_2,
                self.good_platform1_url,
                user_id=self.good_user_2,
            )
            token = m_client.create_egeria_bearer_token(self.good_user_2, "secret")
            start_time = time.perf_counter()
            entity_type = "AssetOwner"

            response = m_client.get_valid_relationship_types(entity_type, output_format="JSON",report_spec="Referenceable")
            duration = time.perf_counter() - start_time

            print(f"\n\tDuration was {duration} seconds")
            if isinstance(response, list | dict):
                print_json("\n\n" + json.dumps(response, indent=4))
            elif type(response) is tuple:
                print(f"Type is {type(response)}")
                print_json("\n\n" + json.dumps(response, indent=4))
            elif type(response) is str:
                print("\n\nGUID is: " + response)
            assert True

        except (
            PyegeriaException
        ) as e:
            print_basic_exception(e)
            assert False, "Invalid request"

        finally:
            m_client.close_session()

    def test_get_valid_classification_types(self):
        try:
            m_client = ValidMetadataManager(
                self.good_view_server_2,
                self.good_platform1_url,
                user_id=self.good_user_2,
            )
            token = m_client.create_egeria_bearer_token(self.good_user_2, "secret")
            start_time = time.perf_counter()
            entity_type = "Asset"

            response = m_client.get_valid_classification_types(entity_type)
            duration = time.perf_counter() - start_time

            print(f"\n\tDuration was {duration} seconds")
            if isinstance(response, list | dict):
                print_json("\n\n" + json.dumps(response, indent=4))
            elif type(response) is tuple:
                print(f"Type is {type(response)}")
                print_json("\n\n" + json.dumps(response, indent=4))
            elif type(response) is str:
                print("\n\nGUID is: " + response)
            assert True

        except PyegeriaException as e:
            print_basic_exception(e)
            assert False, "Invalid request"

        finally:
            m_client.close_session()

    def test_get_typedef_by_name(self):
        try:
            m_client = ValidMetadataManager(
                self.good_view_server_2,
                self.good_platform1_url,
                user_id=self.good_user_2,
            )
            token = m_client.create_egeria_bearer_token(self.good_user_2, "secret")

            start_time = time.perf_counter()
            entity_type = "Project"

            response = m_client.get_typedef_by_name(entity_type)
            duration = time.perf_counter() - start_time

            print(f"\n\tDuration was {duration} seconds")
            if type(response) is dict:
                print_json("\n\n" + json.dumps(response, indent=4))
            elif type(response) is tuple:
                print(f"Type is {type(response)}")
                print_json("\n\n" + json.dumps(response, indent=4))
            elif type(response) is str:
                print("\n\nGUID is: " + response)
            assert True

        except PyegeriaException as e:
            print_basic_exception(e)
            assert False, "Invalid request"

        finally:
            m_client.close_session()

    def test_get_specification_property_types(self):
        try:
            m_client = ValidMetadataManager(
                self.good_view_server_2,
                self.good_platform1_url,
                user_id=self.good_user_2,
            )
            token = m_client.create_egeria_bearer_token(self.good_user_2, "secret")

            start_time = time.perf_counter()

            response = m_client.get_specification_property_types()
            duration = time.perf_counter() - start_time

            print(f"\n\tDuration was {duration} seconds")
            if type(response) is dict:
                print_json("\n\n" + json.dumps(response, indent=4))
            elif type(response) is tuple:
                print(f"Type is {type(response)}")
                print_json("\n\n" + json.dumps(response, indent=4))
            elif type(response) is str:
                print("\n\nGUID is: " + response)
            assert True

        except PyegeriaException as e:
            print_basic_exception(e)
            assert False, "Invalid request"

        finally:
            m_client.close_session()



"""
PDX-License-Identifier: Apache-2.0
Copyright Contributors to the ODPi Egeria project.

    Manage external references to a variety of artifacts.

"""

import asyncio
from typing import Optional

from loguru import logger
from pydantic import HttpUrl

from pyegeria.core._globals import NO_GUID_RETURNED
from pyegeria.view.base_report_formats import select_report_spec, get_report_spec_match
from pyegeria.models import (SearchStringRequestBody, FilterRequestBody, GetRequestBody, NewElementRequestBody,
                             TemplateRequestBody,
                             UpdateElementRequestBody, NewRelationshipRequestBody,
                             DeleteElementRequestBody, DeleteRelationshipRequestBody)
from pyegeria.view.output_formatter import (generate_output,
                                            _extract_referenceable_properties, populate_columns_from_properties,
                                            get_required_relationships)
from pyegeria.core.utils import dynamic_catch

EXTERNAL_REFERENCE_PROPS = ["ExternalReferenceProperties", "ExternalDataSourceProperties",
                            "ExternalModelSourceProperties",
                            "RelatedMediaProperties", "CitedDocumentProperties"]

EXTERNAL_REFERENCE_TYPES = ["ExternalReference", "ExternalDataSource", "ExternalModelSource",
                            "RelatedMedia", "CitedDocument"]
from pyegeria.core._server_client import ServerClient


class ExternalReferences(ServerClient):
    """
    Establish linkage to external references which can be a variety of artifacts. Including media,
    documents, data, and more. Keep in mind that there are several sub-types of external references, each
    with their own property body. This includes:
    
    * ExternalReferenceProperties
    * ExternalDataSourceProperties
    * ExternalModelSourceProperties
    * RelatedMediaProperties
    * CitedDocumentProperties

    Review the documentation on the website for more details.
    
    Attributes:

        server_name: str
            The name of the View Server to connect to.
        platform_url : str
            URL of the server platform to connect to
        user_id : str
            The identity of the user calling the method - this sets a default optionally used by the methods
            when the user doesn't pass the user_id on a method call.
        user_pwd: str
            The password associated with the user_id. Defaults to None
        token: str
            An optional bearer token

    """

    def __init__(self, view_server: str = None, platform_url: str = None, user_id: str = None, user_pwd: Optional[str] = None, token: Optional[str] = None, ):
        ServerClient.__init__(self, view_server, platform_url, user_id, user_pwd, token)
        self.view_server = self.server_name
        self.platform_url = self.platform_url
        self.user_id = self.user_id
        self.user_pwd = self.user_pwd

        self.command_root: str = (
            f"{self.platform_url}/servers/{self.view_server}/api/open-metadata/external-links")

    @dynamic_catch
    async def _async_create_external_reference(self, body: Optional[dict | NewElementRequestBody] = None) -> str:
        """ Create a new generic external_reference. If the body is not present, the display_name, description, category,
            and classification will be used to create a simple, self-anchored external_reference.
            Collections: https://egeria-project.org/concepts/external_reference
            Async version.

        Parameters
        ----------


        body: dict | NewElementRequestBody, optional
            A dict or NewElementRequestBody representing the details of the external_reference to create. If supplied, this
            information will be used to create the external_reference and the other attributes will be ignored. The body is
            validated before being used.

        Returns
        -------
        str - the guid of the created external_reference

        Raises
        ------
        PyegeriaException
            One of the pyegeria exceptions will be raised if there are issues in communications, message format, or
            Egeria errors.
        ValidationError
            Pydantic validation errors are raised if the body does not conform to the NewElementRequestBody.
        NotAuthorizedException
          The principle specified by the user_id does not have authorization for the requested action

        Notes:
        -----
        simple:
        {
          "class": "NewElementRequestBody",
          "isOwnAnchor": true,
          "properties": {
            "class": "ExternalReferenceProperties",
            "qualifiedName": "Must provide a unique name here",
            "name": "Add display name here",
            "description": "Add description of the external_reference here",
            "category": "Add appropriate valid value for type",
            "referenceTitle: "Add reference title here",
            "referenceAbstract": "Add reference abstract here",
            "authors": ["Add author names here"],
            "url": "Add url here",
            "datePublished": "2023-01-01",
            "dateConnected": "2023-01-01",
            "dateCreated": "2023-01-01",
          }
        }

    For related media use this properties body:
        {
          "class": "NewElementRequestBody",
          "isOwnAnchor": true,
          "properties": {
            "class": "RelatedMediaProperties",
            "qualifiedName": "Must provide a unique name here",
            "name": "Add display name here",
            "description": "Add description of the external_reference here",
            "category": "Add appropriate valid value for type",
            "referenceTitle: "Add reference title here",
            "referenceAbstract": "Add reference abstract here",
            "authors": ["Add author names here"],
            "url": "Add url here",
            "datePublished": "2023-01-01",
            "dateConnected": "2023-01-01",
            "dateCreated": "2023-01-01",
            "mediaType": "Add media type here",
            "mediaTypeOtherId": "Add media type other id here",
            "defaultMediaUsage": "Add default media usage here",
            "defaultMediaUsageOtherId": "Add default media usage other id here",
          }
        }

        See https://egeria-project.org/types/0/0015-Linked-Media-Types/ for more information on media types and 
        valid values.

    """

        url = f"{self.command_root}/external-references"
        return await self._async_create_element_body_request(url, EXTERNAL_REFERENCE_PROPS, body)

    @dynamic_catch
    def create_external_reference(self, body: Optional[dict | NewElementRequestBody] = None) -> str:
        """ Create a new generic external_reference. If the body is not present, the display_name, description, category,
             and classification will be used to create a simple, self-anchored external_reference.
             Collections: https://egeria-project.org/concepts/external_reference
             Async version.

             Parameters
             ----------


             body: dict | NewElementRequestBody, optional
                 A dict or NewElementRequestBody representing the details of the external_reference to create. If supplied, this
                 information will be used to create the external_reference and the other attributes will be ignored. The body is
                 validated before being used.

             Returns
             -------
             str - the guid of the created external_reference

             Raises
             ------
             PyegeriaException
                 One of the pyegeria exceptions will be raised if there are issues in communications, message format, or
                 Egeria errors.
             ValidationError
                 Pydantic validation errors are raised if the body does not conform to the NewElementRequestBody.
             NotAuthorizedException
               The principle specified by the user_id does not have authorization for the requested action

             Notes:
             -----
             simple:
             {
               "class": "NewElementRequestBody",
               "isOwnAnchor": true,
               "properties": {
                 "class": "ExternalReferenceProperties",
                 "qualifiedName": "Must provide a unique name here",
                 "name": "Add display name here",
                 "description": "Add description of the external_reference here",
                 "category": "Add appropriate valid value for type",
                 "referenceTitle": "Add reference title here",
                 "referenceAbstract": "Add reference abstract here",
                 "authors": ["Add author names here"],
                 "url": "Add url here",
                 "sources": "Add sources here",
                 "license": "Add license here",
                 "copyright": "Add copyright here",
                 "attribution": "Add attribution here"
               }
             }

              For related media use this properties body:
            {
              "class": "NewElementRequestBody",
              "isOwnAnchor": true,
              "properties": {
                "class": "RelatedMediaProperties",
                "qualifiedName": "Must provide a unique name here",
                "name": "Add display name here",
                "description": "Add description of the external_reference here",
                "category": "Add appropriate valid value for type",
                "referenceTitle": "Add reference title here",
                "referenceAbstract": "Add reference abstract here",
                "authors": ["Add author names here"],
                "url": "Add url here",
                "sources": "Add sources here",
                 "license": "Add license here",
                 "copyright": "Add copyright here",
                 "attribution": "Add attribution here",
                "datePublished": "2023-01-01",
                "dateConnected": "2023-01-01",
                "dateCreated": "2023-01-01",
                "mediaType": "Add media type here",
                "mediaTypeOtherId": "Add media type other id here",
                "defaultMediaUsage": "Add default media usage here",
                "defaultMediaUsageOtherId": "Add default media usage other id here",
              }
            }

            For CitedDocument use this properties body:
            {
              "class": "NewElementRequestBody",
              "isOwnAnchor": true,
              "properties": {
                "class": "CitedDocumentProperties",
                "qualifiedName": "Must provide a unique name here",
                "name": "Add display name here",
                "description": "Add description of the external_reference here",
                "category": "Add appropriate valid value for type",
                "numberOfPates": int,
                "pageRange": "Add page range here",
                "publicationSeries": "Add publication series here",
                "publicationSeriesVolume": "Add publication series volume here",
                "publisher": "Add publisher here",
                "edition": "Add edition here",
                "firstPublicationDate": "2023-01-01",
                "publicationDate": "2023-01-01",
                "publicationCity": "Add publication city here",
                "publicationYear": "publication year",
                "publicationNumbers": ["string"],
                "defaultMediaUsage": "Add default media usage here",
                "defaultMediaUsageOtherId": "Add default media usage other id here",
                "referenceTitle": "Add reference title here",
                "referenceAbstract": "Add reference abstract here",
                "authors": ["Add author names here"],
                "url": "Add url here",
                "sources": "Add sources here",
                 "license": "Add license here",
                 "copyright": "Add copyright here",
                 "attribution": "Add attribution here"
              }
            }

        See https://egeria-project.org/types/0/0015-Linked-Media-Types/ for more information on media types and 
        valid values.


      """

        return asyncio.get_event_loop().run_until_complete(self._async_create_external_reference(body))

    #######

    @dynamic_catch
    async def _async_create_external_reference_from_template(self, body: Optional[dict | TemplateRequestBody] = None) -> str:
        """Create a new metadata element to represent a external_reference using an existing metadata element as a template.
        The template defines additional classifications and relationships that are added to the new external_reference.
        Async version.
    
        Parameters
        ----------
    
        body: dict
            A dict representing the details of the external_reference to create.
    
        Returns
        -------
        str - the guid of the created external_reference
    
        Raises
        ------
        PyegeriaInvalidParameterException
          If the client passes incorrect parameters on the request - such as bad URLs or invalid values
        PyegeriaAPIException
          Raised by the server when an issue arises in processing a valid request
        NotAuthorizedException
          The principle specified by the user_id does not have authorization for the requested action
    
        Notes
        -----
        JSON Structure looks like:
    
        {
          "class": "TemplateRequestBody",
          "anchorGUID": "anchor GUID, if set then isOwnAnchor=false",
          "isOwnAnchor": false,
          "parentGUID": "parent GUID, if set, set all parameters beginning 'parent'",
          "parentRelationshipTypeName": "open metadata type name",
          "parentAtEnd1": true,
          "templateGUID": "template GUID",
          "replacementProperties": {
            "class": "ElementProperties",
            "propertyValueMap" : {
              "propertyName" : {
                "class": "PrimitiveTypePropertyValue",
                "typeName": "string",
                "primitiveTypeCategory" : "OM_PRIMITIVE_TYPE_STRING",
                "primitiveValue" : "value of property"
              }
            }
          },
          "placeholderPropertyValues" : {
            "placeholderProperty1Name" : "property1Value",
            "placeholderProperty2Name" : "property2Value"
          }
        }
    
        """

        if isinstance(body, TemplateRequestBody):
            validated_body = body

        elif isinstance(body, dict):
            validated_body = self._template_request_adapter.validate_python(body)

        url = f"{self.command_root}/external-references/from-template"
        json_body = validated_body.model_dump_json(indent=2, exclude_none=True, by_alias=True)
        logger.info(json_body)
        resp = await self._async_make_request("POST", url, json_body, is_json=True)
        logger.info(f"Create external_reference from template with GUID: {resp.json().get('guid')}")
        return resp.json().get("guid", NO_GUID_RETURNED)

    @dynamic_catch
    def create_external_reference_from_template(self, body: Optional[dict | TemplateRequestBody] = None) -> str:
        """Create a new metadata element to represent a external_reference using an existing metadata element as a template.
        The template defines additional classifications and relationships that are added to the new external_reference.
    
        Parameters
        ----------
        body: dict
            A dict representing the details of the external_reference to create.
    
        Returns
        -------
        str - the guid of the created external_reference
    
        Raises
        ------
        PyegeriaInvalidParameterException
          If the client passes incorrect parameters on the request - such as bad URLs or invalid values
        PyegeriaAPIException
          Raised by the server when an issue arises in processing a valid request
        NotAuthorizedException
          The principle specified by the user_id does not have authorization for the requested action
    
        Notes
        -----
        JSON Structure looks like:
    
        {
          "class": "TemplateRequestBody",
          "anchorGUID": "anchor GUID, if set then isOwnAnchor=false",
          "isOwnAnchor": false,
          "parentGUID": "parent GUID, if set, set all parameters beginning 'parent'",
          "parentRelationshipTypeName": "open metadata type name",
          "parentAtEnd1": true,
          "templateGUID": "template GUID",
          "replacementProperties": {
            "class": "ElementProperties",
            "propertyValueMap" : {
              "propertyName" : {
                "class": "PrimitiveTypePropertyValue",
                "typeName": "string",
                "primitiveTypeCategory" : "OM_PRIMITIVE_TYPE_STRING",
                "primitiveValue" : "value of property"
              }
            }
          },
          "placeholderPropertyValues" : {
            "placeholderProperty1Name" : "property1Value",
            "placeholderProperty2Name" : "property2Value"
          }
        }
        """
        loop = asyncio.get_event_loop()
        resp = loop.run_until_complete(self._async_create_external_reference_from_template(body))
        return resp

        #
        # Manage external_references
        #

    @dynamic_catch
    async def _async_update_external_reference(self, external_reference_guid: str,
                                               body: dict | UpdateElementRequestBody) -> None:
        """ Update the properties of a external_reference. Use the correct properties object (CollectionProperties,
            DigitalProductProperties, AgreementProperties, etc), that is appropriate for your element.
            Collections: https://egeria-project.org/concepts/external_reference
    
            Async version.
        Parameters
        ----------
        external_reference_guid: str
            The guid of the external_reference to update.
    
        body: dict | UpdateElementRequestBody, optional
            A dict or NewElementRequestBody representing the details of the external_reference to create. If supplied, this
            information will be used to create the external_reference and the other attributes will be ignored. The body is
            validated before being used.
        merge_update: bool, optional, default = True
            If true then property changes will be overlaid on top of existing properties. If false, existing
            properties will all be replaced by the set provided in the update request.
    
        Returns
        -------
        None
    
        Raises
        ------
        PyegeriaException
            One of the pyegeria exceptions will be raised if there are issues in communications, message format, or
            Egeria errors.
        ValidationError
            Pydantic validation errors are raised if the body does not conform to the NewElementRequestBody.
        NotAuthorizedException
          The principle specified by the user_id does not have authorization for the requested action
    
        Notes:
        -----
        simple example:
        {
          "class" : "UpdateElementRequestBody",
          "properties": {
            "class" : "CollectionProperties",
            "qualifiedName": "Must provide a unique name here",
            "name" : "Add display name here",
            "description" : "Add description of the external_reference here",
            "category": "Add appropriate valid value for type"
          },
          "externalSourceGUID": "add guid here",
          "externalSourceName": "add qualified name here",
          "effectiveTime" : "{{$isoTimestamp}}",
          "forLineage" : false,
          "forDuplicateProcessing" : false
        }
        """

        # try:

        url = (f"{self.command_root}/external-references/{external_reference_guid}/update")
        await self._async_update_element_body_request(url, EXTERNAL_REFERENCE_PROPS, body)

    @dynamic_catch
    def update_external_reference(self, external_reference_guid: str, body: dict | NewElementRequestBody) -> None:
        """ Update the properties of a external_reference. Use the correct properties object (CollectionProperties,
            DigitalProductProperties, AgreementProperties, etc), that is appropriate for your element.
            Collections: https://egeria-project.org/concepts/external_reference
    
        Parameters
        ----------
        external_reference_guid: str
            The guid of the external_reference to update.
    
        body: dict | NewElementRequestBody, optional
            A dict or NewElementRequestBody representing the details of the external_reference to create. If supplied, this
            information will be used to create the external_reference and the other attributes will be ignored. The body is
            validated before being used.
        merge_update: bool, optional, default = True
            If true then property changes will be overlaid on top of existing properties. If false, existing
            properties will all be replaced by the set provided in the update request.
    
        Returns
        -------
        None
    
        Raises
        ------
        PyegeriaException
            One of the pyegeria exceptions will be raised if there are issues in communications, message format, or
            Egeria errors.
        ValidationError
            Pydantic validation errors are raised if the body does not conform to the NewElementRequestBody.
        NotAuthorizedException
          The principle specified by the user_id does not have authorization for the requested action
    
        Notes:
        -----
        simple example:
        {
          "class" : "UpdateElementRequestBody",
          "properties": {
            "class" : "CollectionProperties",
            "qualifiedName": "Must provide a unique name here",
            "name" : "Add display name here",
            "description" : "Add description of the external_reference here",
            "category": "Add appropriate valid value for type"
          },
          "externalSourceGUID": "add guid here",
          "externalSourceName": "add qualified name here",
          "effectiveTime" : "{{$isoTimestamp}}",
          "forLineage" : false,
          "forDuplicateProcessing" : false
        }
        """

        return asyncio.get_event_loop().run_until_complete(
            self._async_update_external_reference(external_reference_guid, body))

    @dynamic_catch
    async def _async_link_external_reference(self, element_guid: str, ext_ref_guid: str,
                                             body: Optional[dict | NewRelationshipRequestBody] = None) -> None:
        """ Attach an element to an external reference.
            Async version.
    
        Parameters
        ----------
        element_guid: str
            The unique identifier of the element.
        ext_ref_guid: str
            The identifier of the external reference.
        body: dict | NewRelationshipRequestBody, optional, default = None
            A structure representing the details of the relationship.
    
        Returns
        -------
        Nothing
    
        Raises
        ------
        PyegeriaInvalidParameterException
          If the client passes incorrect parameters on the request - such as bad URLs or invalid values
        PyegeriaAPIException
          Raised by the server when an issue arises in processing a valid request
        NotAuthorizedException
          The principle specified by the user_id does not have authorization for the requested action
    
        Notes
        -----
        JSON Structure looks like:
        {
          "class" : "NewRelationshipRequestBody",
          "externalSourceGUID": "add guid here",
          "externalSourceName": "add qualified name here",
          "effectiveTime" : "{{$isoTimestamp}}",
          "forLineage" : false,
          "forDuplicateProcessing" : false,
          "properties": {
            "class": "ExternalReferenceLinkProperties",",
            "label": "add label here",
            "description": "add description here",
            "effectiveFrom": "{{$isoTimestamp}}",
            "effectiveTo": "{{$isoTimestamp}}"
          }
        }

        """

        url = url = (f"{self.command_root}/elements/{element_guid}/external-references/{ext_ref_guid}/attach")
        await self._async_new_relationship_request(url, "ExternalReferenceLinkProperties", body)
        logger.info(f"Linking element {element_guid} to ext. ref.  {ext_ref_guid}")

    @dynamic_catch
    def link_external_reference(self, element_guid: str, ext_ref_guid: str,
                                body: Optional[dict | NewRelationshipRequestBody] = None):
        """ Attach an element to an external reference.
    
            Parameters
            ----------
            element_guid: str
                The unique identifier of the element.
            ext_ref_guid: str
                The identifier of the external reference.
            body: dict | NewRelationshipRequestBody, optional, default = None
                A structure representing the details of the relationship.
    
            Returns
            -------
            Nothing
    
            Raises
            ------
            PyegeriaInvalidParameterException
              If the client passes incorrect parameters on the request - such as bad URLs or invalid values
            PyegeriaAPIException
              Raised by the server when an issue arises in processing a valid request
            NotAuthorizedException
              The principle specified by the user_id does not have authorization for the requested action
    
            Notes
            -----
            JSON Structure looks like:
            {
              "class" : "NewRelationshipRequestBody",
              "externalSourceGUID": "add guid here",
              "externalSourceName": "add qualified name here",
              "effectiveTime" : "{{$isoTimestamp}}",
              "forLineage" : false,
              "forDuplicateProcessing" : false,
              "properties": {
                "class": "ExternalReferenceLinkProperties",",
                "label": "add label here",
                "description": "add description here",
                "effectiveFrom": "{{$isoTimestamp}}",
                "effectiveTo": "{{$isoTimestamp}}"
              }
            }
            """
        loop = asyncio.get_event_loop()
        loop.run_until_complete(self._async_link_external_reference(element_guid, ext_ref_guid, body))

    @dynamic_catch
    async def _async_detach_external_reference(self, element_guid: str, ext_ref_guid: str,
                                               body: Optional[dict | DeleteRelationshipRequestBody] = None) -> None:
        """ Detach an element from an external reference; body is optional. Async version.
    
        Parameters
        ----------
        element_guid: str
            The unique identifier of the subscriber.
        ext_ref_guid: str
            The unique identifier of the subscription.
        body: dict | DeleteRelationshipRequestBody, optional, default = None
            A structure representing the details of the relationship.
    
        Returns
        -------
        Nothing
    
        Raises
        ------
        PyegeriaInvalidParameterException
          If the client passes incorrect parameters on the request - such as bad URLs or invalid values
        PyegeriaAPIException
          Raised by the server when an issue arises in processing a valid request
        NotAuthorizedException
          The principle specified by the user_id does not have authorization for the requested action
    
        Notes
        -----
        JSON Structure looks like:
        {
          "class": "DeleteRelationshipRequestBody",
          "externalSourceGUID": "add guid here",
          "externalSourceName": "add qualified name here",
          "effectiveTime": "{{$isoTimestamp}}",
          "forLineage": false,
          "forDuplicateProcessing": false
        }
        """
        url = (f"{self.command_root}/elements/{element_guid}/external_references/{ext_ref_guid}/detach")

        await self._async_delete_element_request(url, body)
        logger.info(f"Detached element {element_guid} from external reference {ext_ref_guid}")

    def detach_external_reference(self, element_guid: str, ext_ref_guid: str, body: Optional[dict | DeleteRelationshipRequestBody] = None):
        """ Detach an element from an external reference. Request body is optional.
    
        Parameters
        ----------
        element_guid: str
            The unique identifier of the subscriber.
        ext_ref_guid: str
            The unique identifier of the subscription.
        body: dict, optional, default = None
            A dict representing the details of the relationship.
    
        Returns
        -------
        Nothing
    
        Raises
        ------
        PyegeriaInvalidParameterException
          If the client passes incorrect parameters on the request - such as bad URLs or invalid values
        PyegeriaAPIException
          Raised by the server when an issue arises in processing a valid request
        NotAuthorizedException
          The principle specified by the user_id does not have authorization for the requested action
    
        Notes
        -----
        JSON Structure looks like:
        {
          "class": "DeleteRelationshipRequestBody",
          "externalSourceGUID": "add guid here",
          "externalSourceName": "add qualified name here",
          "effectiveTime": "{{$isoTimestamp}}",
          "forLineage": false,
          "forDuplicateProcessing": false
        }
        """
        loop = asyncio.get_event_loop()
        loop.run_until_complete(self._async_detach_external_reference(element_guid, ext_ref_guid, body))

    @dynamic_catch
    async def _async_link_media_reference(self, element_guid: str, media_ref_guid: str,
                                          body: Optional[dict | NewRelationshipRequestBody] = None) -> None:
        """ Attach an element to a related media reference.
            Async version.
    
        Parameters
        ----------
        element_guid: str
            The unique identifier of the element.
        media_ref_guid: str
            The identifier of the external media reference.
        body: dict | NewRelationshipRequestBody, optional, default = None
            A structure representing the details of the relationship.
    
        Returns
        -------
        Nothing
    
        Raises
        ------
        PyegeriaInvalidParameterException
          If the client passes incorrect parameters on the request - such as bad URLs or invalid values
        PyegeriaAPIException
          Raised by the server when an issue arises in processing a valid request
        NotAuthorizedException
          The principle specified by the user_id does not have authorization for the requested action
    
        Notes
        -----
        JSON Structure looks like:
        {
          "class" : "NewRelationshipRequestBody",
          "externalSourceGUID": "add guid here",
          "externalSourceName": "add qualified name here",
          "effectiveTime" : "{{$isoTimestamp}}",
          "forLineage" : false,
          "forDuplicateProcessing" : false,
          "properties": {
            "class": "MediaReferenceProperties",
            "mediaId": "",
            "description": "",
            "mediaUsage": "ICON",
            "mediaUsageOtherId": "",
            "effectiveFrom": "{{$isoTimestamp}}",
            "effectiveTo": "{{$isoTimestamp}}"
          }
        }

        """
        url = f"{self.command_root}/elements/{element_guid}/media-references/{media_ref_guid}/attach"
        await self._async_new_relationship_request(url, "MediaReferenceProperties", body)
        logger.info(f"Linking element {element_guid} to media reference  {media_ref_guid}")

    @dynamic_catch
    def link_media_reference(self, element_guid: str, media_ref_guid: str,
                             body: Optional[dict | NewRelationshipRequestBody] = None):
        """ Attach an element to an external media reference.
    
            Parameters
            ----------
            element_guid: str
                The unique identifier of the element.
            media_ref_guid: str
                The identifier of the external reference.
            body: dict | NewRelationshipRequestBody, optional, default = None
                A structure representing the details of the relationship.
    
            Returns
            -------
            Nothing
    
            Raises
            ------
            PyegeriaInvalidParameterException
              If the client passes incorrect parameters on the request - such as bad URLs or invalid values
            PyegeriaAPIException
              Raised by the server when an issue arises in processing a valid request
            NotAuthorizedException
              The principle specified by the user_id does not have authorization for the requested action
    
            Notes
            -----
            JSON Structure looks like:
            {
              "class" : "NewRelationshipRequestBody",
              "externalSourceGUID": "add guid here",
              "externalSourceName": "add qualified name here",
              "effectiveTime" : "{{$isoTimestamp}}",
              "forLineage" : false,
              "forDuplicateProcessing" : false,
              "properties": {
                "class": "MediaReferenceProperties",
                "mediaId": "",
                "description": "",
                "mediaUsage": "ICON",
                "mediaUsageOtherId": "",
                "effectiveFrom": "{{$isoTimestamp}}",
                "effectiveTo": "{{$isoTimestamp}}"
              }
            }
            """
        loop = asyncio.get_event_loop()
        loop.run_until_complete(self._async_link_media_reference(element_guid, media_ref_guid, body))

    @dynamic_catch
    async def _async_detach_media_reference(self, element_guid: str, media_ref_guid: str,
                                            body: Optional[dict | DeleteRelationshipRequestBody] = None) -> None:
        """ Detach an element from an external media reference; body is optional. Async version.
    
        Parameters
        ----------
        element_guid: str
            The unique identifier of the subscriber.
        media_ref_guid: str
            The unique identifier of the subscription.
        body: dict | DeleteRelationshipRequestBody, optional, default = None
            A structure representing the details of the relationship.
    
        Returns
        -------
        Nothing
    
        Raises
        ------
        PyegeriaInvalidParameterException
          If the client passes incorrect parameters on the request - such as bad URLs or invalid values
        PyegeriaAPIException
          Raised by the server when an issue arises in processing a valid request
        NotAuthorizedException
          The principle specified by the user_id does not have authorization for the requested action
    
        Notes
        -----
        JSON Structure looks like:
        {
          "class": "DeleteRelationshipRequestBody",
          "externalSourceGUID": "add guid here",
          "externalSourceName": "add qualified name here",
          "effectiveTime": "{{$isoTimestamp}}",
          "forLineage": false,
          "forDuplicateProcessing": false
        }
        """
        url = (
            f"{self.command_root}/elements/{element_guid}/media-references/{media_ref_guid}/detach")

        await self._async_delete_relationship_request(url, body)
        logger.info(f"Detached element {element_guid} from external media reference {media_ref_guid}")

    @dynamic_catch
    def detach_media_reference(self, element_guid: str, media_ref_guid: str, body: Optional[dict | DeleteRelationshipRequestBody] = None):
        """ Detach an element from an external media reference. Request body is optional.
    
        Parameters
        ----------
        element_guid: str
            The unique identifier of the subscriber.
        media_ref_guid: str
            The unique identifier of the subscription.
        body: dict, optional, default = None
            A dict representing the details of the relationship.
    
        Returns
        -------
        Nothing
    
        Raises
        ------
        PyegeriaInvalidParameterException
          If the client passes incorrect parameters on the request - such as bad URLs or invalid values
        PyegeriaAPIException
          Raised by the server when an issue arises in processing a valid request
        NotAuthorizedException
          The principle specified by the user_id does not have authorization for the requested action
    
        Notes
        -----
        JSON Structure looks like:
        {
          "class": "DeleteRelationshipRequestBody",
          "externalSourceGUID": "add guid here",
          "externalSourceName": "add qualified name here",
          "effectiveTime": "{{$isoTimestamp}}",
          "forLineage": false,
          "forDuplicateProcessing": false
        }
        """
        loop = asyncio.get_event_loop()
        loop.run_until_complete(self._async_detach_media_reference(element_guid, media_ref_guid, body))

    @dynamic_catch
    async def _async_link_cited_document(self, element_guid: str, cited_doc_guid: str,
                                         body: Optional[dict | NewRelationshipRequestBody] = None) -> None:
        """ Attach an element to a cited document reference.
            Async version.
    
        Parameters
        ----------
        element_guid: str
            The unique identifier of the element.
        cited_doc_guid: str
            The identifier of the external cited document reference.
        body: dict | NewRelationshipRequestBody, optional, default = None
            A structure representing the details of the relationship.
    
        Returns
        -------
        Nothing
    
        Raises
        ------
        PyegeriaInvalidParameterException
          If the client passes incorrect parameters on the request - such as bad URLs or invalid values
        PyegeriaAPIException
          Raised by the server when an issue arises in processing a valid request
        NotAuthorizedException
          The principle specified by the user_id does not have authorization for the requested action
    
        Notes
        -----
        JSON Structure looks like:
        {
          "class" : "NewRelationshipRequestBody",
          "externalSourceGUID": "add guid here",
          "externalSourceName": "add qualified name here",
          "effectiveTime" : "{{$isoTimestamp}}",
          "forLineage" : false,
          "forDuplicateProcessing" : false,
          "properties": {
                    "class": "CitedDocumentLinkProperties",
                    "referenceId": "add reference id here",
                    "description": "Add description of the external_reference here",
                    "pages": "Add pages here"
                  }
        }
        """
        url = f"{self.command_root}/elements/{element_guid}/cited-document-references/{cited_doc_guid}/attach"
        await self._async_new_relationship_request(url, "CitedDocumentLinkProperties", body)
        logger.info(f"Linking element {element_guid} to cited document  {cited_doc_guid}")

    @dynamic_catch
    def link_cited_document(self, element_guid: str, cited_doc_guid: str,
                            body: Optional[dict | NewRelationshipRequestBody] = None):
        """ Attach an element to an external media reference.
    
            Parameters
            ----------
            element_guid: str
                The unique identifier of the element.
            cited_doc_guid: str
                The identifier of the external reference.
            body: dict | NewRelationshipRequestBody, optional, default = None
                A structure representing the details of the relationship.
    
            Returns
            -------
            Nothing
    
            Raises
            ------
            PyegeriaInvalidParameterException
              If the client passes incorrect parameters on the request - such as bad URLs or invalid values
            PyegeriaAPIException
              Raised by the server when an issue arises in processing a valid request
            NotAuthorizedException
              The principle specified by the user_id does not have authorization for the requested action
    
            Notes
            -----
            JSON Structure looks like:
            {
              "class" : "NewRelationshipRequestBody",
              "externalSourceGUID": "add guid here",
              "externalSourceName": "add qualified name here",
              "effectiveTime" : "{{$isoTimestamp}}",
              "forLineage" : false,
              "forDuplicateProcessing" : false,
              "properties": {
                    "class": "CitedDocumentLinkProperties",
                    "referenceId": "add reference id here",
                    "description": "Add description of the external_reference here",
                    "pages": "Add pages here"
                  }
             }
            """
        loop = asyncio.get_event_loop()
        loop.run_until_complete(self._async_link_cited_document(element_guid, cited_doc_guid, body))

    @dynamic_catch
    async def _async_detach_cited_document(self, element_guid: str, cited_doc_guid: str,
                                           body: Optional[dict | DeleteRelationshipRequestBody] = None) -> None:
        """ Detach an element from an cited document reference; body is optional. Async version.
    
        Parameters
        ----------
        element_guid: str
            The unique identifier of the subscriber.
        cited_doc_guid: str
            The unique identifier of the subscription.
        body: dict | DeleteRelationshipRequestBody, optional, default = None
            A structure representing the details of the relationship.
    
        Returns
        -------
        Nothing
    
        Raises
        ------
        PyegeriaInvalidParameterException
          If the client passes incorrect parameters on the request - such as bad URLs or invalid values
        PyegeriaAPIException
          Raised by the server when an issue arises in processing a valid request
        NotAuthorizedException
          The principle specified by the user_id does not have authorization for the requested action
    
        Notes
        -----
        JSON Structure looks like:
        {
          "class": "DeleteRelationshipRequestBody",
          "externalSourceGUID": "add guid here",
          "externalSourceName": "add qualified name here",
          "effectiveTime": "{{$isoTimestamp}}",
          "forLineage": false,
          "forDuplicateProcessing": false
        }
        """
        url = f"{self.command_root}/elements/{element_guid}/cited-document-references/{cited_doc_guid}/detach"

        await self._async_delete_relationship_request(url, body)
        logger.info(f"Detached element {element_guid} from cited document reference {cited_doc_guid}")

    @dynamic_catch
    def detach_cited_document(self, element_guid: str, cited_doc_guid: str, body: Optional[dict | DeleteRelationshipRequestBody] = None):
        """ Detach an element from acited document reference. Request body is optional.
    
        Parameters
        ----------
        element_guid: str
            The unique identifier of the subscriber.
        cited_doc_guid: str
            The unique identifier of the subscription.
        body: dict, optional, default = None
            A dict representing the details of the relationship.
    
        Returns
        -------
        Nothing
    
        Raises
        ------
        PyegeriaInvalidParameterException
          If the client passes incorrect parameters on the request - such as bad URLs or invalid values
        PyegeriaAPIException
          Raised by the server when an issue arises in processing a valid request
        NotAuthorizedException
          The principle specified by the user_id does not have authorization for the requested action
    
        Notes
        -----
        JSON Structure looks like:
        {
          "class": "DeleteRelationshipRequestBody",
          "externalSourceGUID": "add guid here",
          "externalSourceName": "add qualified name here",
          "effectiveTime": "{{$isoTimestamp}}",
          "forLineage": false,
          "forDuplicateProcessing": false
        }
        """
        loop = asyncio.get_event_loop()
        loop.run_until_complete(self._async_detach_cited_document(element_guid, cited_doc_guid, body))

    #
    # do deletes etc
    #
    @dynamic_catch
    async def _async_delete_external_reference(self, ext_ref_guid: str,
                                               body: Optional[dict | DeleteRelationshipRequestBody] = None,
                                               cascade: bool = False) -> None:
        """ Delete an external reference. Async Version.
    
        Parameters
        ----------
        ext_ref_guid: str
            The guid of the governance definition to delete.
    
        cascade: bool, optional, defaults to True
            If true, a cascade delete is performed.
    
        body: dict DeleteRelationshipRequestBodyt, optional, default = None
            A dict representing the details of the relationship.
    
        Returns
        -------
        Nothing
    
        Raises
        ------
        PyegeriaInvalidParameterException
          If the client passes incorrect parameters on the request - such as bad URLs or invalid values
        ValidationException
          Raised by pydantic when the request body is invalid.
        NotAuthorizedException
          The principle specified by the user_id does not have authorization for the requested action
    
        Notes
        _____
        JSON Structure looks like:
        {
          "class" : "DeleteRelationshipRequestBody",
          "externalSourceGUID": "add guid here",
          "externalSourceName": "add qualified name here",
          "effectiveTime" : "{{$isoTimestamp}}",
          "forLineage" : false,
          "forDuplicateProcessing" : false
        }
        """
        url = f"{self.command_root}/external-references/{ext_ref_guid}/delete"

        await self._async_delete_element_request(url, body, cascade)
        logger.info(f"Deleted collection {ext_ref_guid} with cascade {cascade}")

    @dynamic_catch
    def delete_external_reference(self, ext_ref_guid: str, body: Optional[dict | DeleteElementRequestBody] = None,
                                  cascade: bool = False) -> None:
        """Delete an external reference..
    
        Parameters
        ----------
        ext_ref_guid: str
            The guid of the external reference to delete.
    
        cascade: bool, optional, defaults to True
            If true, a cascade delete is performed.
    
        body: dict DeleteElementRequestBody, optional, default = None
            A dict representing the details of the relationship.
    
        Returns
        -------
        Nothing
    
        Raises
        ------
        PyegeriaInvalidParameterException
          If the client passes incorrect parameters on the request - such as bad URLs or invalid values
        ValidationException
          Raised by pydantic when the request body is invalid.
        NotAuthorizedException
          The principle specified by the user_id does not have authorization for the requested action
    
        Notes
        _____
        JSON Structure looks like:
        {
          "class" : "DeleteElementRequestBody",
          "externalSourceGUID": "add guid here",
          "externalSourceName": "add qualified name here",
          "effectiveTime" : "{{$isoTimestamp}}",
          "forLineage" : false,
          "forDuplicateProcessing" : false
        }
        """
        loop = asyncio.get_event_loop()
        loop.run_until_complete(self._async_delete_external_reference(ext_ref_guid, body, cascade))

    @dynamic_catch
    async def _async_find_external_references(
        self,
        search_string: str = "*",
        body: Optional[dict | SearchStringRequestBody] = None,
        starts_with: bool = True,
        ends_with: bool = False,
        ignore_case: bool = False,
        start_from: int = 0,
        page_size: int = 100,
        output_format: str = 'JSON',
        report_spec: str | dict = "ExternalReference",
        **kwargs
    ) -> list | str:
        """ Retrieve the list of external reference metadata elements that contain the search string. Async Version.

        Parameters
        ----------
        search_string: str
            Search string to match against - None or '*' indicate match against all external references.
        starts_with : bool, [default=True], optional
            Starts with the supplied string.
        ends_with : bool, [default=False], optional
            Ends with the supplied string
        ignore_case : bool, [default=False], optional
            Ignore case when searching
        anchor_domain: str, optional
            The anchor domain to search in.
        metadata_element_type: str, optional
            The type of metadata element to search for.
        metadata_element_subtypes: list[str], optional
            The subtypes of metadata element to search for.
        skip_relationships: list[str], optional
            The types of relationships to skip.
        include_only_relationships: list[str], optional
            The types of relationships to include.
        skip_classified_elements: list[str], optional
            The types of classified elements to skip.
        include_only_classified_elements: list[str], optional
            The types of classified elements to include.
        graph_query_depth: int, [default=3], optional
            The depth of the graph query.
        governance_zone_filter: list[str], optional
            The governance zones to search in.
        as_of_time: str, optional
            The time to search as of.
        effective_time: str, optional
            The effective time to search at.
        relationship_page_size: int, [default=0], optional
            The page size for relationships.
        limit_results_by_status: list[str], optional
            The statuses to limit results by.
        sequencing_order: str, optional
            The order to sequence results by.
        sequencing_property: str, optional
            The property to sequence results by.
        output_format: str, default = "JSON"
            - one of "MD", "LIST", "FORM", "REPORT", "DICT", "MERMAID" or "JSON"
        report_spec: str | dict , optional, default = "ExternalReference"
            - The desired output columns/fields to include.
        start_from: int, [default=0], optional
            When multiple pages of results are available, the page number to start from.
        page_size: int, [default=100]
            The number of items to return in a single page.
        property_names: list[str], optional
            The names of properties to search for.
        body: dict | SearchStringRequestBody, optional, default = None
            - if provided, the search parameters in the body will supercede other attributes, such as "search_string"

        Returns
        -------
        List | str

        Output depends on the output format specified.

        Raises
        ------

        ValidationError
          If the client passes incorrect parameters on the request that don't conform to the data model.
        PyegeriaException
          Issues raised in communicating or server side processing.
        NotAuthorizedException
          The principle specified by the user_id does not have authorization for the requested action

        """
        url = str(HttpUrl(f"{self.command_root}/external-references/by-search-string"))

        # Merge explicit parameters with kwargs
        params = {
            'search_string': search_string,
            'body': body,
            'starts_with': starts_with,
            'ends_with': ends_with,
            'ignore_case': ignore_case,
            'start_from': start_from,
            'page_size': page_size,
            'output_format': output_format,
            'report_spec': report_spec
        }
        params.update(kwargs)

        # Filter out None values, but keep search_string even if None (it's required)
        params = {k: v for k, v in params.items() if v is not None or k == 'search_string'}

        response = await self._async_find_request(url, _type="ExternalReference",
                                                  _gen_output=self._generate_external_reference_output, **params)

        return response

    @dynamic_catch
    def find_external_references(
        self,
        search_string: str = "*",
        body: Optional[dict | SearchStringRequestBody] = None,
        starts_with: bool = True,
        ends_with: bool = False,
        ignore_case: bool = False,
        start_from: int = 0,
        page_size: int = 100,
        output_format: str = 'JSON',
        report_spec: str | dict = "ExternalReference",
        **kwargs
    ) -> list | str:
        """ Retrieve the list of external reference metadata elements that contain the search string.

        Parameters
        ----------
        search_string: str
            Search string to match against - None or '*' indicate match against all external references.
        starts_with : bool, [default=True], optional
            Starts with the supplied string.
        ends_with : bool, [default=False], optional
            Ends with the supplied string
        ignore_case : bool, [default=False], optional
            Ignore case when searching
        anchor_domain: str, optional
            The anchor domain to search in.
        metadata_element_type: str, optional
            The type of metadata element to search for.
        metadata_element_subtypes: list[str], optional
            The subtypes of metadata element to search for.
        skip_relationships: list[str], optional
            The types of relationships to skip.
        include_only_relationships: list[str], optional
            The types of relationships to include.
        skip_classified_elements: list[str], optional
            The types of classified elements to skip.
        include_only_classified_elements: list[str], optional
            The types of classified elements to include.
        graph_query_depth: int, [default=3], optional
            The depth of the graph query.
        governance_zone_filter: list[str], optional
            The governance zones to search in.
        as_of_time: str, optional
            The time to search as of.
        effective_time: str, optional
            The effective time to search at.
        relationship_page_size: int, [default=0], optional
            The page size for relationships.
        limit_results_by_status: list[str], optional
            The statuses to limit results by.
        sequencing_order: str, optional
            The order to sequence results by.
        sequencing_property: str, optional
            The property to sequence results by.
        output_format: str, default = "JSON"
            - one of "MD", "LIST", "FORM", "REPORT", "DICT", "MERMAID" or "JSON"
        report_spec: str | dict , optional, default = "ExternalReference"
            - The desired output columns/fields to include.
        start_from: int, [default=0], optional
            When multiple pages of results are available, the page number to start from.
        page_size: int, [default=100]
            The number of items to return in a single page.
        property_names: list[str], optional
            The names of properties to search for.
        body: dict | SearchStringRequestBody, optional, default = None
            - if provided, the search parameters in the body will supercede other attributes, such as "search_string"

        Returns
        -------
        List | str

        Output depends on the output format specified.

        Raises
-------

        ValidationError
          If the client passes incorrect parameters on the request that don't conform to the data model.
        PyegeriaException
          Issues raised in communicating or server side processing.
        NotAuthorizedException
          The principle specified by the user_id does not have authorization for the requested action

        """
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(
            self._async_find_external_references(
                search_string=search_string,
                body=body,
                starts_with=starts_with,
                ends_with=ends_with,
                ignore_case=ignore_case,
                start_from=start_from,
                page_size=page_size,
                output_format=output_format,
                report_spec=report_spec,
                **kwargs
            )
        )

    @dynamic_catch
    async def _async_get_external_references_by_name(self, filter_string: Optional[str] = None,
                                                     classification_names: Optional[list[str]] = None,
                                                     body: Optional[dict | FilterRequestBody] = None,
                                                     start_from: int = 0, page_size: int = 0,
                                                     output_format: str = 'JSON',
                                                     report_spec: str | dict = "ExternalReference") -> list | str:
        """ Returns the list of external references with a particular name.

            Parameters
            ----------
            filter_string: str,
                name to use to find matching collections.
            classification_names: list[str], optional, default = None
                type of collection to filter by - e.g., DataDict, Folder, Root
            body: dict, optional, default = None
                Provides, a full request body. If specified, the body supercedes the name parameter.
            start_from: int, [default=0], optional
                        When multiple pages of results are available, the page number to start from.
            page_size: int, [default=None]
                The number of items to return in a single page. If not specified, the default will be taken from
                the class instance.
            output_format: str, default = "JSON"
                - one of "DICT", "MERMAID" or "JSON"
            report_spec: dict , optional, default = None
                The desired output columns/fields to include.

            Returns
            -------
            List | str

            A list of collections match matching the name. Returns a string if none found.

            Raises
            ------

            PyegeriaInvalidParameterException
              If the client passes incorrect parameters on the request - such as bad URLs or invalid values
            PyegeriaAPIException
              Raised by the server when an issue arises in processing a valid request
            NotAuthorizedException
              The principle specified by the user_id does not have authorization for the requested action
        """
        url = str(HttpUrl(f"{self.command_root}/external-references/by-name"))
        response = await self._async_get_name_request(url, _type="ExternalReference",
                                                      _gen_output=self._generate_external_reference_output,
                                                      filter_string=filter_string,
                                                      classification_names=classification_names, start_from=start_from,
                                                      page_size=page_size, output_format=output_format,
                                                      report_spec=report_spec, body=body)

        return response

    def get_external_references_by_name(self, filter_string: Optional[str] = None, classification_names: Optional[list[str]] = None,
                                        body: Optional[dict | FilterRequestBody] = None,
                                        start_from: int = 0, page_size: int = 0, output_format: str = 'JSON',
                                        report_spec: str | dict = "ExternalReference") -> list | str:
        """Returns the list of external references matching the filter string. Async version.
            The search string is located in the request body and is interpreted as a plain string.
            The request parameters, startsWith, endsWith, and ignoreCase can be used to allow a fuzzy search.

        Parameters
        ----------
        filter_string: str,
            name to use to find matching collections.
        classification_names: list[str], optional, default = None
            type of collection to filter by - e.g., DataDict, Folder, Root
        body: dict, optional, default = None
            Provides, a full request body. If specified, the body supercedes the name parameter.
        start_from: int, [default=0], optional
                    When multiple pages of results are available, the page number to start from.
        page_size: int, [default=None]
            The number of items to return in a single page. If not specified, the default will be taken from
            the class instance.
        output_format: str, default = "JSON"
            - one of "DICT", "MERMAID" or "JSON"
         report_spec: str | dict, optional, default = None
                The desired output columns/fields to include.

        Returns
        -------
        List | str

        A list of collections match matching the search string. Returns a string if none found.

        Raises
        ------
        PyegeriaException

        """
        return asyncio.get_event_loop().run_until_complete(
            self._async_get_external_references_by_name(filter_string, classification_names, body, start_from,
                                                        page_size,
                                                        output_format, report_spec))

    @dynamic_catch
    async def _async_get_external_reference_by_guid(self, ext_ref_guid: str, element_type: Optional[str] = None,
                                                    body: Optional[dict | GetRequestBody] = None,
                                                    output_format: str = 'JSON',
                                                    report_spec: str | dict = "ExternalReference") -> dict | str:
        """Return the properties of a specific external reference. Async version.

        Parameters
        ----------
        ext_ref_guid: str,
            unique identifier of the external reference to retrieve.
        element_type: str, default = None, optional
             type of externak reference ExternalReference, RelatedMedia, etc.
        body: dict | GetRequestBody, optional, default = None
            full request body.
        output_format: str, default = "JSON"
            - one of "DICT", "MERMAID" or "JSON"
         report_spec: str | dict, optional, default = None
                The desired output columns/fields to include.

        Returns
        -------
        dict | str

        A JSON dict representing the specified collection. Returns a string if none found.

        Raises
        ------

        PyegeriaInvalidParameterException
          If the client passes incorrect parameters on the request - such as bad URLs or invalid values
        PyegeriaAPIException
          Raised by the server when an issue arises in processing a valid request
        NotAuthorizedException
          The principle specified by the user_id does not have authorization for the requested action

        Notes
        ----
        Body sample:
        {
          "class": "GetRequestBody",
          "asOfTime": "{{$isoTimestamp}}",
          "effectiveTime": "{{$isoTimestamp}}",
          "forLineage": false,
          "forDuplicateProcessing": false
        }
        """

        url = str(HttpUrl(f"{self.command_root}/external-references/{ext_ref_guid}/retrieve"))
        type = element_type if element_type else "ExternalReference"

        response = await self._async_get_guid_request(url, _type=type,
                                                      _gen_output=self._generate_external_reference_output,
                                                      output_format=output_format, report_spec=report_spec,
                                                      body=body)

        return response

    @dynamic_catch
    def get_external_reference_by_guid(self, ext_ref_guid: str, element_type: Optional[str] = None,
                                       body: Optional[dict | GetRequestBody] = None,
                                       output_format: str = 'JSON',
                                       report_spec: str | dict = "ExternalReference") -> dict | str:
        """ Return the properties of a specific external reference. Async version.

            Parameters
            ----------
            ext_ref_guid: str,
                unique identifier of the external reference to retrieve.
            element_type: str, default = None, optional
                type of element - ExternalReference, RelatedMedia, etc.
            body: dict | GetRequestBody, optional, default = None
                full request body.
            output_format: str, default = "JSON"
                - one of "DICT", "MERMAID" or "JSON"
            report_spec: dict , optional, default = None
                The desired output columns/fields to include.


            Returns
            -------
            dict | str

            A JSON dict representing the specified collection. Returns a string if none found.

            Raises
            ------

            PyegeriaInvalidParameterException
              If the client passes incorrect parameters on the request - such as bad URLs or invalid values
            PyegeriaAPIException
              Raised by the server when an issue arises in processing a valid request
            NotAuthorizedException
              The principle specified by the user_id does not have authorization for the requested action

            Notes
            ----
            Body sample:
            {
              "class": "AnyTimeRequestBody",
              "asOfTime": "{{$isoTimestamp}}",
              "effectiveTime": "{{$isoTimestamp}}",
              "forLineage": false,
              "forDuplicateProcessing": false
            }
        """
        return asyncio.get_event_loop().run_until_complete(
            self._async_get_external_reference_by_guid(ext_ref_guid, element_type, body,
                                                       output_format, report_spec))

    def _extract_external_reference_properties(self, element: dict, columns_struct: dict) -> dict:
        """
        Extract common properties from a external_reference element and populate into the provided columns_struct.

        Args:
            element (dict): The external_reference element
            columns_struct (dict): The columns structure to populate

        Returns:
            dict: columns_struct with column 'value' fields populated
        """
        # First, populate from element.properties using the utility
        col_data = populate_columns_from_properties(element, columns_struct)

        columns_list = col_data.get("formats", {}).get("attributes", [])

        # Populate header-derived values
        header_props = _extract_referenceable_properties(element)
        for column in columns_list:
            key = column.get('key')
            if key in header_props:
                column['value'] = header_props.get(key)
            elif isinstance(key, str) and key.lower() == 'guid':
                column['value'] = header_props.get('GUID')

        # Derived/computed fields
        # external_referenceCategories are classifications
        classification_names = ""
        classifications = element.get('elementHeader', {}).get("external_referenceCategories", [])
        for classification in classifications:
            classification_names += f"{classification['classificationName']}, "
        if classification_names:
            for column in columns_list:
                if column.get('key') == 'classifications':
                    column['value'] = classification_names[:-2]
                    break

        # Populate requested relationship-based columns generically from top-level keys
        col_data = get_required_relationships(element, col_data)

        # Subject area classification
        subject_area = element.get('elementHeader', {}).get("subjectArea", "") or ""
        subj_val = ""
        if isinstance(subject_area, dict):
            subj_val = subject_area.get("classificationProperties", {}).get("subjectAreaName", "")
        for column in columns_list:
            if column.get('key') == 'subject_area':
                column['value'] = subj_val
                break

        # Mermaid graph
        mermaid_val = element.get('mermaidGraph', "") or ""
        for column in columns_list:
            if column.get('key') == 'mermaid':
                column['value'] = mermaid_val
                break

        logger.trace(f"Extracted/Populated columns: {col_data}")

        return col_data

    def _generate_external_reference_output(self, elements: dict | list[dict], filter_string: Optional[str] = None,
                                            element_type_name: Optional[str] = None, output_format: str = "DICT",
                                            report_spec: dict | str = None, **kwargs) -> str | list[dict]:
        """ Generate output for external_references in the specified format.

            Args:
                elements (Union[Dict, List[Dict]]): Dictionary or list of dictionaries containing data field elements
                filter_string (Optional[str]): The search string used to find the elements
                element_type_name (Optional[str]): The type of external_reference
                output_format (str): The desired output format (MD, FORM, REPORT, LIST, DICT, MERMAID, HTML)
                report_spec (Optional[dict], optional): List of dictionaries containing column data. Defaults
                to None.
                **kwargs: Additional arguments.

            Returns:
                Union[str, List[Dict]]: Formatted output as a string or list of dictionaries
        """
        return self._generate_formatted_output(
            elements=elements,
            query_string=filter_string,
            entity_type=element_type_name or "ExternalReference",
            output_format=output_format,
            extract_properties_func=self._extract_external_reference_properties,
            report_spec=report_spec,
            **kwargs
        )


from typing import Union, Dict, List, Optional

if __name__ == "__main__":
    print("Main-Collection Manager")



#!/usr/bin/env python3
"""
SPDX-Lic
ense-Identifier: Apache-2.0
Copyright Contributors to the ODPi Egeria project.

Unit tests for the Utils helper functions using the Pytest framework.


A simple display for glossary terms
"""
import argparse
import json
import os
import sys
import time
from typing import Optional

# Add the repository root to sys.path to ensure local pyegeria is used
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "..")))

from rich import box, print
from rich.console import Console
from rich.prompt import Prompt
from rich.table import Table

from pyegeria import (
    PyegeriaException,
    ValidMetadataManager, print_basic_exception,

)
from pyegeria.core.config import load_app_config, get_app_config

# Legacy environment variables - now handled via pyegeria.core.config
# EGERIA_METADATA_STORE = os.environ.get("EGERIA_METADATA_STORE", "active-metadata-store")
# ... (others removed to use unified config)


def display_metadata_values(
    property_name: str,
    type_name: str,
    server: str,
    url: str,
    username: str,
    user_pass: str,
    save_output: bool,
    jupyter: bool = False,
    width: int = 200,
):
    try:
        m_client = ValidMetadataManager(server, url, user_id=username)
        token = m_client.create_egeria_bearer_token(username, user_pass)
    except PyegeriaException as e:
        print_basic_exception(e)
        return
    except ValueError as ve:
        print(f"[red]Input error: {ve}[/red]")
        return

    def generate_table(property_name: str, type_name: Optional[str]) -> Table:
        """Make a new table."""
        table = Table(
            title=f"Valid Metadata Values for Property: {property_name} of type {type_name} @ {time.asctime()}",
            style="bold bright_white on black",
            row_styles=["bold bright_white on black"],
            header_style="white on dark_blue",
            title_style="bold white on black",
            caption_style="white on black",
            show_lines=True,
            box=box.ROUNDED,
            caption=f"Valid Metadata Values for Server '{server}' @ Platform - {url}",
            expand=True,
        )

        table.add_column("Category")
        table.add_column("Display Name")

        table.add_column("Preferred Value")
        table.add_column("Deprecated")
        table.add_column("Case Sensitive")
        table.add_column("Additional Properties")
        table.add_column("Description")

        valid_values = m_client.get_valid_metadata_values(property_name, type_name)

        if valid_values is None:
            name = " "
            type_name = " "
            created = " "
            priority = " "
            due = " "
            completed = " "
            status = " "
            sponsor = " "
        elif type(valid_values[0]) == str:
            raise ValueError(
                "-->This is not a known metadata property with a valid value"
            )
        else:
            for value in valid_values:
                category = value.get("category", "None")
                display_name = value.get("displayName", "None")
                preferred_value = value.get("preferredValue", "None ")
                deprecated = str(value.get("isDeprecated", "None "))
                case_sensitive = str(value.get("isCaseSensitive", "None"))
                description = value.get("description", "None")
                additional_properties = value.get("additionalProperties", "None")
                if additional_properties is not None:
                    props = json.dumps(additional_properties)
                table.add_row(
                    category,
                    display_name,
                    preferred_value,
                    deprecated,
                    case_sensitive,
                    props,
                    description,
                )

        m_client.close_session()
        return table

    try:
        # with Live(generate_table(), refresh_per_second=4, screen=True) as live:
        #     while True:
        #         time.sleep(2)
        #         live.update(generate_table())
        console = Console(width=width, force_terminal=not jupyter, record=True)
        with console.pager(styles=True):
            console.print(generate_table(property_name, type_name))
        if save_output:
            console.save_html("valid-metadata-values.html")

    except (
        PyegeriaException,
        ValueError,
    ) as e:
        if type(e) is str:
            print(e)
        else:
            print_basic_exception(e)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--server", help="Name of the server to display status for")
    parser.add_argument("--url", help="URL Platform to connect to")
    parser.add_argument("--userid", help="User Id")
    parser.add_argument("--save-output", help="Save output to file?")
    parser.add_argument("--password", help="User Password")
    # parser.add_argument("--sponsor", help="Name of sponsor to search")
    args = parser.parse_args()

    config = load_app_config()
    user_profile = config.User_Profile
    env = config.Environment

    server = args.server if args.server is not None else env.egeria_view_server
    url = args.url if args.url is not None else env.egeria_platform_url
    userid = args.userid if args.userid is not None else user_profile.user_name
    user_pass = args.password if args.password is not None else user_profile.user_pwd
    save_output = args.save_output if args.save_output is not None else False

    try:
        property_name = Prompt.ask(
            "Enter the Property to retrieve:", default="projectHealth"
        )
        type_name = Prompt.ask(
            "Enter the Metadata Type to filter on (Optional, use :: or leave blank for None):", default="::"
        )
        if type_name == "":
            type_name = None

        display_metadata_values(
            property_name, type_name, server, url, userid, user_pass, save_output, 
            jupyter=env.egeria_jupyter, width=env.console_width
        )
    except PyegeriaException as e:
        print_basic_exception(e)
    except ValueError as ve:
        print(f"[red]Value Error: {ve}[/red]")
    except KeyboardInterrupt:
        pass
    except Exception as e:
        print(f"[red]An unexpected error occurred: {e}[/red]")


if __name__ == "__main__":
    main()



"""
SPDX-License-Identifier: Apache-2.0
Copyright Contributors to the ODPi Egeria project.
"""

import inspect
import json
from json import JSONDecodeError

import validators

from pyegeria.core._exceptions import PyegeriaInvalidParameterException

"""
This package contains internally used validators.

"""


def validate_user_id(user_id: str) -> bool:
    """
    Validate that the provided user id is neither null nor empty.

    Parameters
    ----------
    user_id : str  The user id string to validate

    Returns
    -------
    bool: True if valid, If invalid input, a PyegeriaAuthenticationException is raised.

    Raises
    ------
    PyegeriaAuthenticationException
        If the provided user id is null or empty
    """
    if (user_id is None) or len(user_id) == 0:
        context: dict = {}
        context['calling_frame'] = inspect.currentframe().f_back
        context['caller_method'] = inspect.currentframe().f_back.f_code.co_name
        additional_info = {
            "reason": "Invalid user name - its empty",
            "userid": user_id,
            }
        raise PyegeriaInvalidParameterException(None,context, additional_info)
    else:
        return True


def validate_server_name(server_name: str) -> bool:
    """
    Validate that the provided server name is neither null nor empty.

    Parameters
    ----------
    server_name : str  The user id string to validate

    Returns
    -------
    bool - True if valid, If invalid an PyegeriaInvalidParameterException is raised.

    Raises
    ------
    PyegeriaInvalidParameterException
        If the provided server name is null or empty

    """

    if (server_name is None) or (len(server_name) == 0):
        context: dict = {}
        context['calling_frame'] = inspect.currentframe().f_back
        context['caller_method'] = inspect.currentframe().f_back.f_code.co_name
        additional_info = {
            "reason": "Invalid server name - its empty", "input_parameters": f"server_name={server_name}",
            }
        raise PyegeriaInvalidParameterException(None, context, additional_info)
    else:
        return True


def validate_guid(guid: str) -> bool:
    """
    Validate that the provided guid is neither null nor empty.

    Parameters
    ----------
    guid : str  The user id string to validate

    Returns
    -------
    bool - True if valid, If invalid an PyegeriaInvalidParameterException is raised.

    Raises
    ------
    PyegeriaInvalidParameterException
        If the provided guid is null or empty
    """
    if (guid is None) or (len(guid) == 0) or (type(guid) is not str):
        context: dict = {}
        context['calling_frame'] = inspect.currentframe().f_back
        context['caller_method'] = inspect.currentframe().f_back.f_code.co_name
        additional_info = {
            "reason": "Invalid GUID", "input_parameters": f"guid = {guid}"
            }
        raise PyegeriaInvalidParameterException(None,context, additional_info)
    else:
        return True


def validate_name(name: str) -> bool:
    """
    Validate that the provided name is neither null nor empty.

    Parameters
    ----------
    name: str  The user id string to validate

    Returns
    -------
    bool - True if valid, If invalid an PyegeriaInvalidParameterException is raised.

    Raises
    ------
    PyegeriaInvalidParameterException
        If the provided name is null or empty
    """

    if (name is None) or (len(name) == 0):
        context: dict = {}
        context['calling_frame'] = inspect.currentframe().f_back
        context['caller_method'] = inspect.currentframe().f_back.f_code.co_name
        additional_info = {
            "reason": "Invalid `name`", "input_parameters": f"name = {name}"
            }
        raise PyegeriaInvalidParameterException(None,context, additional_info)
    else:
        return True


def validate_search_string(search_string: str) -> bool:
    """
    Validate that the provided search string is neither null nor empty.

    Parameters
    ----------
    search_string : str  The user id string to validate

    Returns
    -------
    bool - True if valid, If invalid an PyegeriaInvalidParameterException is raised.

    Raises
    ------
    PyegeriaInvalidParameterException
        If the provided search string is null or empty
    """

    if (search_string is None) or (len(search_string) == 0):
        context: dict = {}
        context['calling_frame'] = inspect.currentframe().f_back
        context['caller_method'] = inspect.currentframe().f_back.f_code.co_name
        additional_info = {
            "reason": "Invalid `name`", "input_parameters": f"search_string={search_string}"
            }
        raise PyegeriaInvalidParameterException(None,context, additional_info)
    else:
        return True


def validate_public(is_public: bool) -> bool:
    """
    Validate that the provided flag is boolean.

    Parameters
    ----------
    is_public : bool  The flag must be boolean

    Returns
    -------
    bool - True if valid, If invalid an PyegeriaInvalidParameterException is raised.

    Raises
    ------
    PyegeriaInvalidParameterException
        If the provided public flag is null or empty
    """

    if is_public is None:
        context: dict = {}
        context['calling_frame'] = inspect.currentframe().f_back
        context['caller_method'] = inspect.currentframe().f_back.f_code.co_name
        additional_info = {
            "reason": "Invalid `name`", "input_parameters": f"is_public={is_public}"
            }
        raise PyegeriaInvalidParameterException(None,context, additional_info)
    else:
        return True


def validate_url(url: str) -> bool:
    """
    Validate that the provided url is neither null nor empty. The syntax of the url
    string is also checked to see that it conforms to standards.

    Parameters
    ----------
    url : str  The url string to validate.

    Returns
    -------
    bool - True if valid, If invalid an PyegeriaInvalidParameterException is raised.

    """
    context: dict = {}
    context['calling_frame'] = inspect.currentframe().f_back
    context['caller_method'] = inspect.currentframe().f_back.f_code.co_name

    if (url is None) or (len(url) == 0):

        additional_info = {"reason": "The provided URL is invalid - it is empty",
                           "input_parameters": f"URL: {url}"}
        raise PyegeriaInvalidParameterException(None,context, additional_info)

    # The following hack allows localhost to be used as a hostname - which is disallowed by the
    # validations package
    if ("localhost" in url) and ("localhost." not in url):
        url = url.replace("localhost", "127.0.0.1")

    result = validators.url(url)
    if result is not True:
        additional_info = {
            "reason": "The provided URL is invalid",
            "input_parameters": f"URL: {url}"
            }
        raise PyegeriaInvalidParameterException(None,context, additional_info)
    else:
        return True


def is_json(txt: str) -> bool:
    """
    Parameters
    ----------
    txt : str
        The string to check if it is a valid JSON.

    Returns
    -------
    bool
        True if the string is a valid JSON, False otherwise.
    """
    try:
        json.loads(txt)
        return True
    except (ValueError, JSONDecodeError) as e:
        print(e)
        return False



"""
SPDX-License-Identifier: Apache-2.0
Copyright Contributors to the ODPi Egeria project.

EgeriaCat: Catalog-oriented client using multiple inheritance.

This client combines methods from:
- AssetCatalog (asset_catalog_omvs)
- GlossaryManager (glossary_manager)
- ProjectManager (project_manager)
- MyProfile (my_profile_omvs)

to provide a single, convenient surface for catalog users.
"""
from pyegeria.omvs.asset_catalog import AssetCatalog
from pyegeria.omvs.glossary_manager import GlossaryManager
from pyegeria.omvs.project_manager import ProjectManager
from pyegeria.omvs.my_profile import MyProfile


class EgeriaCat(ProjectManager, GlossaryManager, AssetCatalog, MyProfile):
    """
    Catalog-oriented client that exposes the combined methods of
    ProjectManager, GlossaryManager, AssetCatalog, and MyProfile via inheritance.

    Notes:
    - ProjectManager is listed first to prefer the BaseServerClient (ServerClient) token/session
      implementations in method resolution order.
    - Each parent class is explicitly initialized to ensure its internal state
      (URLs, roots, etc.) is set up.
    """

    def __init__(
        self,
        view_server: str = None,
        platform_url: str = None,
        user_id: str = None,
        user_pwd: str = None,
        token: str = None,
    ):
        # Initialize each parent explicitly (do not rely on cooperative super)
        ProjectManager.__init__(self, view_server, platform_url, user_id, user_pwd, token)
        GlossaryManager.__init__(self, view_server, platform_url, user_id, user_pwd, token)
        AssetCatalog.__init__(self, view_server, platform_url, user_id, user_pwd, token)
        MyProfile.__init__(self, view_server, platform_url, user_id, user_pwd, token)


if __name__ == "__main__":
    print("Egeria Cat Client")



"""
SPDX-License-Identifier: Apache-2.0
Copyright Contributors to the ODPi Egeria project.

This class is the overall Egeria Client covering all of the functional pyegeria classes. May not be appropriate
for all use cases..using the more role based clients is often appropriate:
    * EgeriaCat - for catalog users
    * EgeriaConfig - for configuring the Egeria platform and servers
    * EgeriaMy - for personal actions
    * EgeriaOps - for operations and administration
    * EgeriaTech - for technical users such as data scientists and engineers

"""
from typing import Optional, Dict, Any

from pyegeria.egeria_config_client import EgeriaConfig
from pyegeria.omvs.action_author import ActionAuthor
from pyegeria.omvs.actor_manager import ActorManager
from pyegeria.omvs.asset_catalog import AssetCatalog
from pyegeria.omvs.asset_maker import AssetMaker
from pyegeria.omvs.automated_curation import AutomatedCuration
from pyegeria.omvs.classification_explorer import ClassificationExplorer
from pyegeria.omvs.collection_manager import CollectionManager
from pyegeria.omvs.community_matters_omvs import CommunityMatters
from pyegeria.omvs.data_designer import DataDesigner
from pyegeria.omvs.data_discovery import DataDiscovery
from pyegeria.omvs.data_engineer import DataEngineer
from pyegeria.omvs.digital_business import DigitalBusiness
from pyegeria.omvs.external_links import ExternalReferences
from pyegeria.omvs.feedback_manager import FeedbackManager
from pyegeria.omvs.full_omag_server_config import FullServerConfig
from pyegeria.omvs.glossary_manager import GlossaryManager
from pyegeria.omvs.governance_officer import GovernanceOfficer
from pyegeria.omvs.lineage_linker import LineageLinker
from pyegeria.omvs.location_arena import LocationArena
from pyegeria.omvs.metadata_expert import MetadataExpert
from pyegeria.omvs.metadata_explorer_omvs import MetadataExplorer
from pyegeria.omvs.my_profile import MyProfile
from pyegeria.omvs.notification_manager import NotificationManager
from pyegeria.omvs.people_organizer import PeopleOrganizer
from pyegeria.omvs.product_manager import ProductManager
from pyegeria.omvs.project_manager import ProjectManager
from pyegeria.omvs.reference_data import ReferenceDataManager
from pyegeria.omvs.registered_info import RegisteredInfo
from pyegeria.omvs.runtime_manager import RuntimeManager
from pyegeria.omvs.schema_maker import SchemaMaker
from pyegeria.omvs.server_operations import ServerOps
from pyegeria.omvs.solution_architect import SolutionArchitect
from pyegeria.omvs.specification_properties import SpecificationProperties
from pyegeria.omvs.subject_area import SubjectArea
from pyegeria.omvs.template_manager_omvs import TemplateManager
from pyegeria.omvs.time_keeper import TimeKeeper
from pyegeria.omvs.valid_metadata import ValidMetadataManager
from pyegeria.omvs.valid_metadata_lists import ValidMetadataLists
from pyegeria.omvs.valid_type_lists import ValidTypeLists


class Egeria:
    """
    Overall Egeria client that composes all functional pyegeria clients using lazy loading.
    """

    def __init__(
        self,
        view_server: str = None,
        platform_url: str = None,
        user_id: str = None,
        user_pwd: str = None,
        token: str = None,
    ):
        self.view_server = view_server
        self.platform_url = platform_url
        self.user_id = user_id
        self.user_pwd = user_pwd
        self.token = token

        self._subclient_map = {
            "action_author": ActionAuthor,
            "actor_manager": ActorManager,
            "asset_catalog": AssetCatalog,
            "asset_maker": AssetMaker,
            "auto_curate": AutomatedCuration,
            "automated_curation": AutomatedCuration,
            "class_mgr": ClassificationExplorer,
            "classification_manager": ClassificationExplorer,
            "collections": CollectionManager,
            "collection_manager": CollectionManager,
            "community": CommunityMatters,
            "community_matters": CommunityMatters,
            "config": EgeriaConfig,
            "data_discovery": DataDiscovery,
            "data_engineer": DataEngineer,
            "designer": DataDesigner,
            "data_designer": DataDesigner,
            "digital_business": DigitalBusiness,
            "expert": MetadataExpert,
            "metadata_expert": MetadataExpert,
            "explorer": MetadataExplorer,
            "metadata_explorer": MetadataExplorer,
            "external_refs": ExternalReferences,
            "external_references": ExternalReferences,
            "feedback": FeedbackManager,
            "feedback_manager": FeedbackManager,
            "full_server_config": FullServerConfig,
            "glossary": GlossaryManager,
            "glossary_manager": GlossaryManager,
            "gov_officer": GovernanceOfficer,
            "governance_officer": GovernanceOfficer,
            "lineage_linker": LineageLinker,
            "location": LocationArena,
            "my_profile": MyProfile,
            "notifications": NotificationManager,
            "notification_manager": NotificationManager,
            "people_organizer": PeopleOrganizer,
            "product_manager": ProductManager,
            "projects": ProjectManager,
            "project_manager": ProjectManager,
            "reference_data": ReferenceDataManager,
            "reg_info": RegisteredInfo,
            "registered_info": RegisteredInfo,
            "runtime": RuntimeManager,
            "runtime_manager": RuntimeManager,
            "schema_maker": SchemaMaker,
            "server_ops": ServerOps,
            "sol_arch": SolutionArchitect,
            "solution_architect": SolutionArchitect,
            "specification_properties": SpecificationProperties,
            "subject_area": SubjectArea,
            "subject_area_manager": SubjectArea,
            "templates": TemplateManager,
            "template_manager": TemplateManager,
            "time_keeper": TimeKeeper,
            "valid": ValidMetadataManager,
            "valid_metadata": ValidMetadataManager,
            "valid_metadata_lists": ValidMetadataLists,
            "valid_type_lists": ValidTypeLists,
        }
        self._instantiated_clients = {}

    def _get_subclient(self, attr_name: str):
        if attr_name not in self._instantiated_clients:
            client_cls = self._subclient_map[attr_name]
            self._instantiated_clients[attr_name] = client_cls(
                self.view_server, self.platform_url, self.user_id, self.user_pwd, self.token
            )
        return self._instantiated_clients[attr_name]

    def __getattr__(self, name):
        """Resolves attributes via instantiated or mapped subclients"""
        # Allow direct access to sub-clients if the name matches a key in the map
        if name in self._subclient_map:
            return self._get_subclient(name)

        for inst in self._instantiated_clients.values():
            if hasattr(inst, name):
                return getattr(inst, name)
        for attr_name, client_cls in self._subclient_map.items():
            if hasattr(client_cls, name):
                return getattr(self._get_subclient(attr_name), name)
        raise AttributeError(f"{self.__class__.__name__} object has no attribute {name}")

    def create_egeria_bearer_token(self, user_id: str = None, user_pwd: str = None):
        # Use server_ops as a reliable helper for token generation
        helper = self._get_subclient("server_ops")
        token_val = helper.create_egeria_bearer_token(user_id, user_pwd)
        self.set_bearer_token(token_val)
        return token_val

    def set_bearer_token(self, token: str) -> None:
        self.token = token
        for sub in self._instantiated_clients.values():
            sub.set_bearer_token(token)

    def get_token(self) -> str:
        if self.token: return self.token
        for sub in self._instantiated_clients.values():
            if hasattr(sub, "get_token"):
                return sub.get_token()
        return None

    def close_session(self) -> None:
        for sub in self._instantiated_clients.values():
            if hasattr(sub, "close_session"):
                try: sub.close_session()
                except Exception: pass
        self._instantiated_clients.clear()


"""
SPDX-License-Identifier: Apache-2.0

Thin adapter helpers to surface pyegeria FormatSets as MCP-style tools without
side effects. This does NOT implement an MCP transport/server; it focuses on
programmatic functions that an MCP server entry point can call.

Only format sets that advertise DICT or ALL are considered eligible.
"""
from __future__ import annotations

import json
import sys
from typing import Any, Dict, Optional

from loguru import logger

from pyegeria.view.base_report_formats import (
    list_mcp_format_sets,
    select_report_spec, find_report_specs,
)
from pyegeria.egeria_tech_client import EgeriaTech
from pyegeria.view.format_set_executor import exec_report_spec, _async_run_report


def list_reports() -> dict:
    """List eligible format sets as MCP tools (support DICT or ALL)."""
    return  list_mcp_format_sets()

def run_find_report_specs(perspective: str= None, question: str= None, report_spec:str= None) -> Dict[str, Any]:
    """
    Find report specs that match the given perspective and question.

    Args:
        perspective (str): The perspective to search for (e.g., "Data Steward").
        question (str): The question to search for (e.g., "What is the current status of the project?").
        report_spec (str): The report spec to search for (e.g., "ProjectStatusReport").

    Returns:
        list[dict]: A list of dictionaries, each representing a matching report spec item.
    """
    perspective = None if perspective == "*" else perspective
    question = None if question == "*" else question
    report_spec = None if report_spec == "*" else report_spec

    report_specs = find_report_specs(perspective=perspective, question=question, report_spec=report_spec)
    if not report_specs:
        raise ValueError(f"No report specs found for perspective '{perspective}' and question '{question}'")
    return {"Matching Report Specs" : report_specs}

def describe_report(name: str, output_type: str = "DICT") -> Dict[str, Any]:
    """
    Describe a format set for MCP discovery. If outputType != ANY, a concrete format
    will be resolved; otherwise only metadata/action are returned.
    """
    meta = select_report_spec(name, output_type)
    if not meta:
        raise ValueError(f"Unknown or incompatible format set: {name}")
    return meta


def _execute_egeria_call_blocking(
        *,
        report: str,
        params: Optional[Dict[str, Any]] = None,
        view_server: Optional[str] = None,
        view_url: Optional[str] = None,
        user: Optional[str] = None,
        user_pass: Optional[str] = None,) -> Dict[str, Any]:
    """
    Executes the synchronous, blocking Egeria client call on a dedicated worker thread.

    You must replace the hardcoded return with your actual Egeria client logic here.
    All code in this function runs in a blocking, synchronous manner.
    """

    print(
        f"Format set: {report}\nparams: {json.dumps(params)}\nview_server: {view_server}\nview_url: {view_url}\nuser: {user}\nuser_pass: {user_pass}",
        file=sys.stderr)
    # Lazy import of settings to avoid circulars when optional args are None
    # from pyegeria.config import settings as _settings
    from pyegeria.core.config import settings as _settings


    return exec_report_spec(
        format_set_name=report,
        output_format="DICT",
        params=params or {},
        view_server=view_server if view_server is not None else _settings.Environment.egeria_view_server,
        view_url=view_url if view_url is not None else _settings.Environment.egeria_view_server_url,
        user=user if user is not None else _settings.User_Profile.user_name,
        user_pass=user_pass if user_pass is not None else _settings.User_Profile.user_pwd,
    )
    # # Returning the hardcoded success for now to prove the async structure works.
    # return {
    #     "status": "SUCCESS (Thread Test)",
    #     "report_name": report_name,
    #     "message": "The blocking call was successfully run on a separate thread, preventing timeout."
    # }



def run_report(
    *,
    report: str,
    params: Optional[Dict[str, Any]] = None,
    view_server: Optional[str] = None,
    view_url: Optional[str] = None,
    user: Optional[str] = None,
    user_pass: Optional[str] = None,
) -> Dict[str, Any]:
    """
    Execute a format set action as an MCP-style tool. Enforces DICT/ALL by default.
    Caller may pass credentials explicitly; otherwise defaults are used from config.
    """
    print(f"Format set: {report}\nparams: {json.dumps(params)}\nview_server: {view_server}\nview_url: {view_url}\nuser: {user}\nuser_pass: {user_pass}", file=sys.stderr)
    # Lazy import of settings to avoid circulars when optional args are None
    from pyegeria.core.config import settings as _settings
    logger.info(f"Format set: {report}\nparams: {json.dumps(params)}\nview_server: {view_server}\nview_url: {view_url}\nuser: {user}\nuser_pass: {user_pass}")
    return exec_report_spec(
        format_set_name=report,
        output_format="DICT",
        params=params or {},
        view_server=view_server if view_server is not None else _settings.Environment.egeria_view_server,
        view_url=view_url if view_url is not None else _settings.Environment.egeria_view_server_url,
        user=user if user is not None else _settings.User_Profile.user_name,
        user_pass=user_pass if user_pass is not None else _settings.User_Profile.user_pwd,
    )

async def _async_run_report_tool(
    *,
    report: str,
    egeria_client: EgeriaTech,
    params: Optional[Dict[str, Any]] = None,
    output_format: str = "DICT"
) -> Dict[str, Any]:
    """
    Execute a format set action as an MCP-style tool. Enforces DICT/ALL by default.
    Caller may pass credentials explicitly; otherwise defaults are used from config.
    """
    # Lazy import of settings to avoid circulars when optional args are None

    print(f"Report: {report}\n params: {json.dumps(params)}\n", file=sys.stderr)
    result = await _async_run_report(report, egeria_client, output_format=output_format, params=params)
    return result



"""
PDX-License-Identifier: Apache-2.0
Copyright Contributors to the ODPi Egeria project.

Scenario tests for LineageLinker.
"""
import time
from typing import List, Optional
from dataclasses import dataclass

from rich import print as rprint
from rich.console import Console
from rich.table import Table

from pyegeria.omvs.lineage_linker import LineageLinker
from pyegeria.omvs.asset_maker import AssetMaker
from pyegeria.core._exceptions import PyegeriaException, PyegeriaTimeoutException

VIEW_SERVER = "qs-view-server"
PLATFORM_URL = "https://localhost:9443"
USER_ID = "peterprofile"
USER_PWD = "secret"

console = Console(width=200)


@dataclass
class TestResult:
    scenario_name: str
    status: str
    duration: float
    message: str = ""


class LineageLinkerScenarioTester:
    def __init__(self):
        self.client: Optional[LineageLinker] = None
        self.asset_maker: Optional[AssetMaker] = None
        self.results: List[TestResult] = []
        self.created_elements: List[str] = []

    def setup(self) -> bool:
        try:
            self.client = LineageLinker(VIEW_SERVER, PLATFORM_URL, USER_ID, USER_PWD)
            self.asset_maker = AssetMaker(VIEW_SERVER, PLATFORM_URL, USER_ID, USER_PWD)
            self.client.create_egeria_bearer_token(USER_ID, USER_PWD)
            self.asset_maker.create_egeria_bearer_token(USER_ID, USER_PWD)
            return True
        except Exception:
            return False

    def teardown(self):
        if self.client:
            self.client.close_session()
        if self.asset_maker:
            self.asset_maker.close_session()

    def scenario_lineage_link_lifecycle(self) -> TestResult:
        name = "Lineage Link Lifecycle"
        start_time = time.perf_counter()
        try:
            # 1. Create two assets to link
            guid1 = self.asset_maker.create_csv_asset("SourceAsset", "Source for lineage", "source.csv")
            guid2 = self.asset_maker.create_csv_asset("TargetAsset", "Target for lineage", "target.csv")
            self.created_elements.extend([guid1, guid2])

            # 2. Link them with Lineage
            body = {
                "class": "NewRelationshipRequestBody",
                "properties": {
                    "class": "DataFlowProperties",
                    "label": "Scenario Flow",
                }
            }
            rel_guid = self.client.link_lineage(guid1, "DataFlow", guid2, body)
            
            return TestResult(name, "PASSED", time.perf_counter() - start_time, f"Linked {guid1} to {guid2}")
        except PyegeriaTimeoutException as e:
            rprint(f"[bold yellow]Timeout in {name}; continuing.[/bold yellow]")
            return TestResult(name, "WARNING", time.perf_counter() - start_time, f"Timeout: {e}")
        except PyegeriaException as e:
            return TestResult(name, "FAILED", time.perf_counter() - start_time, str(e))
        except Exception as e:
            return TestResult(name, "FAILED", time.perf_counter() - start_time, f"Unexpected error: {e}")

    def run_all_scenarios(self):
        if not self.setup():
            rprint("[bold red]Setup failed, skipping scenarios[/bold red]")
            return
        
        self.results.append(self.scenario_lineage_link_lifecycle())
        self.teardown()
        self.print_results_summary()

    def print_results_summary(self):
        table = Table(title="Lineage Linker Scenario Results")
        table.add_column("Scenario", style="cyan")
        table.add_column("Status", justify="center")
        table.add_column("Duration (s)", justify="right")
        table.add_column("Message")
        for res in self.results:
            if res.status == "PASSED":
                style = "green"
            elif res.status == "WARNING":
                style = "yellow"
            else:
                style = "red"
            table.add_row(res.scenario_name, f"[{style}]{res.status}[/{style}]", f"{res.duration:.2f}", res.message)
        console.print(table)


def test_lineage_linker_scenarios():
    tester = LineageLinkerScenarioTester()
    tester.run_all_scenarios()


if __name__ == "__main__":
    test_lineage_linker_scenarios()



#!/usr/bin/env python3
"""
SPDX-License-Identifier: Apache-2.0
Copyright Contributors to the ODPi Egeria project.

A simple viewer for collections - provide the root and we display the hierarchy

"""
import argparse
import os
import sys

from rich import print
from rich.console import Console
from rich.markdown import Markdown
from rich.panel import Panel
from rich.prompt import Prompt
from rich.text import Text
from rich.tree import Tree

from pyegeria import (
    AutomatedCuration,
    PyegeriaException,
    print_basic_exception,
    settings,
    config_logging
)

EGERIA_USER = os.environ.get("EGERIA_USER", "erinoverview")
EGERIA_USER_PASSWORD = os.environ.get("EGERIA_USER_PASSWORD", "secret")

app_config = settings.Environment
config_logging()
console = Console(width = app_config.console_width)



def tech_details_viewer(
    tech: str,
    server_name: str,
    platform_url: str,
    user: str,
    user_password: str,
    jupyter: bool = app_config.egeria_jupyter,
    width: int = app_config.console_width,
):
    console = Console(width=width, force_terminal=not jupyter)

    def view_tech_details(
        a_client: AutomatedCuration, root_collection_name: str, tree: Tree
    ) -> Tree:
        l2: Tree = None
        tech_details = a_client.get_tech_type_detail(tech)
        if (type(tech_details) is dict) and (len(tech_details) > 0):
            name = tech_details.get("displayName", "---")
            qualified_name = tech_details.get("qualifiedName", "---")
            category = tech_details.get("category", "---")
            description = tech_details.get("description", "---")

            style = "bold bright_white on black"

            bas_md = f"* **Name**: {name}\n"
            bas_md += f"* **QualifiedName**: {qualified_name}\n"
            bas_md += f"* **Category**: {category}\n"
            bas_md += f"* **Technology Description**: {description}\n"

            bas_out = Markdown(bas_md)

            l2 = tree.add(Panel(bas_out))
            ext_ref = tech_details.get("externalReferences", None)

            if ext_ref is not None:
                uri = ext_ref[0]["properties"].get("uri", "---")
                # console.print(f" {type(ext_ref)}, {len(ext_ref)}")
                l2 = tree.add(Panel(Markdown(f"* URI: {uri}", style)))

            resource_list = tech_details.get("resourceList", None)
            if resource_list:
                t_r = tree.add(Panel("Resource List[bold red]"))
                for resource in resource_list:
                    resource_use = Text(
                        f"[bold bright_white]{resource.get('resourceUse', '---')}"
                    )
                    resource_use_description = Text(
                        f"[bold bright_white]{resource.get('description', '---')}"
                    )
                    type_name = Text(
                        f"[bold bright_white]{resource['relatedElement']['elementHeader']['type'].get('typeName', '---')}"
                    )
                    unique_name = Text(
                        f"[bold bright_white]{resource['relatedElement']['properties'].get('qualifiedName', '---')}"
                    )
                    related_guid = Text(
                        f"[bold bright_white]{resource['relatedElement']['elementHeader'].get('guid', '---')}"
                    )
                    resource_text = (
                        f"[bold red]Resource\n"
                        f"[white]Resource use: {resource_use}[white]\nDescription: "
                        f"{resource_use_description}\n[white]Type Name: {type_name}\n"
                        f"[white]Unique Name: {unique_name}\n[white]Related GUID: {related_guid}\n"
                    )
                    p = Panel.fit(resource_text)
                    tt = t_r.add(p, style=style)
            else:
                tt = tree.add(Panel(f"Resource details for {tech} where not found"))

        else:
            tt = tree.add(
                Panel(
                    f"Tech type {tech} was not found - please check the tech type name"
                )
            )

        return tt

    try:
        tree = Tree(
            f"{tech}",
            style="bold bright_white on black",
            guide_style="bold bright_blue",
        )
        a_client = AutomatedCuration(server_name, platform_url, user_id=user)

        token = a_client.create_egeria_bearer_token(user, user_password)
        view_tech_details(a_client, tech, tree)
        print(tree)

    except (
        PyegeriaException,
    ) as e:
        print_basic_exception(e)


def main():
    parser = argparse.ArgumentParser()

    parser.add_argument("--server", help="Name of the server to display status for")
    parser.add_argument("--url", help="URL Platform to connect to")
    parser.add_argument("--userid", help="User Id")
    parser.add_argument("--password", help="User Password")
    args = parser.parse_args()

    server = args.server if args.server is not None else app_config.egeria_view_server
    url = args.url if args.url is not None else app_config.egeria_view_server_url
    userid = args.userid if args.userid is not None else EGERIA_USER
    user_pass = args.password if args.password is not None else EGERIA_USER_PASSWORD
    try:
        tech = Prompt.ask(
            "Enter the Technology to start from:", default="PostgreSQL Server"
        )
        tech_details_viewer(tech, server, url, userid, user_pass)
    except KeyboardInterrupt:
        pass


if __name__ == "__main__":
    main()



"""
SPDX-License-Identifier: Apache-2.0
Copyright Contributors to the ODPi Egeria project.

This module is for testing the FeedbackManager class and methods.
The routines assume that pytest is being used as the test tool and framework.

A running Egeria environment is needed to run these tests.
"""

import json
import time

from pydantic import ValidationError

from pyegeria import (
    PyegeriaException,
    print_exception_table,
    print_validation_error,
    print_basic_exception,
)
from pyegeria.core._exceptions import (
    PyegeriaInvalidParameterException,
    PyegeriaAPIException,
    PyegeriaUnauthorizedException,
)
from pyegeria.omvs.feedback_manager import FeedbackManager

disable_ssl_warnings = True


class TestFeedbackManager:
    good_platform1_url = "https://localhost:9443"
    good_platform2_url = "https://oak.local:9443"
    bad_platform1_url = "https://localhost:9443"

    good_user_1 = "garygeeke"
    good_user_2 = "erinoverview"
    good_user_3 = "peterprofile"
    bad_user_1 = "eviledna"
    bad_user_2 = ""
    good_user_2_pwd = "secret"
    good_server_1 = "simple-metadata-store"
    good_server_2 = "qs-view-server"
    good_server_3 = "active-metadata-store"
    good_server_4 = "integration-daemon"
    good_server_5 = "fluffy_kv"
    good_server_6 = "cocoVIew1"
    good_engine_host_1 = "governDL01"
    good_view_server_1 = "view-server"
    good_view_server_2 = "qs-view-server"
    bad_server_1 = "coco"
    bad_server_2 = ""
    test_element_guid = "49bc1002-1b0a-4194-9305-3607c713726d"

    def test_add_like_to_element(self):
        """Test adding a like to an element"""
        try:
            f_client = FeedbackManager(
                self.good_view_server_2,
                self.good_platform1_url,
                user_id=self.good_user_2,
                user_pwd=self.good_user_2_pwd,
            )

            token = f_client.create_egeria_bearer_token(
                self.good_user_2, self.good_user_2_pwd
            )
            
            # Use a known element GUID from your test environment
            # This should be replaced with an actual test element
            # test_element_guid = "test-element-guid-123"
            
            start_time = time.perf_counter()
            response = f_client.add_like_to_element(
                self.test_element_guid,

            )
            duration = time.perf_counter() - start_time
            
            print(f"\n\tDuration was {duration} seconds")
            print(f"\n\tResponse was: {json.dumps(response, indent=2)}")
            
            assert True, "Like added successfully"

        except PyegeriaInvalidParameterException as e:
            print_exception_table(e)
            assert False, "Invalid parameter exception"
        except PyegeriaAPIException as e:
            print_exception_table(e)
            assert False, "API exception"
        except PyegeriaUnauthorizedException as e:
            print_exception_table(e)
            assert False, "Unauthorized exception"
        except ValidationError as e:
            print_validation_error(e)
            assert False, "Validation error"
        except Exception as e:
            print_basic_exception(e)
            assert False, "Unexpected exception"
        finally:
            f_client.close_session()

    def test_add_rating_to_element(self):
        """Test adding a rating to an element"""
        try:
            f_client = FeedbackManager(
                self.good_view_server_2,
                self.good_platform1_url,
                user_id=self.good_user_2,
                user_pwd=self.good_user_2_pwd,
            )

            token = f_client.create_egeria_bearer_token(
                self.good_user_2, self.good_user_2_pwd
            )
            
            # test_element_guid = "test-element-guid-123"
            
            # Rating body with star rating and review
            rating_body = {
                "starRating": 5,
                "review": "Excellent element!"
            }
            
            start_time = time.perf_counter()
            response = f_client.add_rating_to_element(
                self.test_element_guid,
                is_public=True,
                body=rating_body
            )
            duration = time.perf_counter() - start_time
            
            print(f"\n\tDuration was {duration} seconds")
            print(f"\n\tResponse was: {json.dumps(response, indent=2)}")
            
            assert True, "Rating added successfully"

        except PyegeriaInvalidParameterException as e:
            print_exception_table(e)
            assert False, "Invalid parameter exception"
        except PyegeriaAPIException as e:
            print_exception_table(e)
            assert False, "API exception"
        except PyegeriaUnauthorizedException as e:
            print_exception_table(e)
            assert False, "Unauthorized exception"
        except ValidationError as e:
            print_validation_error(e)
            assert False, "Validation error"
        except Exception as e:
            print_basic_exception(e)
            assert False, "Unexpected exception"
        finally:
            f_client.close_session()

    def test_get_attached_likes(self):
        """Test retrieving likes attached to an element"""
        try:
            f_client = FeedbackManager(
                self.good_view_server_2,
                self.good_platform1_url,
                user_id=self.good_user_2,
                user_pwd=self.good_user_2_pwd,
            )

            token = f_client.create_egeria_bearer_token(
                self.good_user_2, self.good_user_2_pwd
            )
            
            test_element_guid = "test-element-guid-123"
            
            start_time = time.perf_counter()
            response = f_client.get_attached_likes(
                test_element_guid,
                start_from=0,
                page_size=50
            )
            duration = time.perf_counter() - start_time
            
            print(f"\n\tDuration was {duration} seconds")
            print(f"\n\tResponse was: {json.dumps(response, indent=2)}")
            
            assert True, "Likes retrieved successfully"

        except PyegeriaInvalidParameterException as e:
            print_exception_table(e)
            assert False, "Invalid parameter exception"
        except PyegeriaAPIException as e:
            print_exception_table(e)
            assert False, "API exception"
        except PyegeriaUnauthorizedException as e:
            print_exception_table(e)
            assert False, "Unauthorized exception"
        except ValidationError as e:
            print_validation_error(e)
            assert False, "Validation error"
        except Exception as e:
            print_basic_exception(e)
            assert False, "Unexpected exception"
        finally:
            f_client.close_session()

    def test_get_attached_ratings(self):
        """Test retrieving ratings attached to an element"""
        try:
            f_client = FeedbackManager(
                self.good_view_server_2,
                self.good_platform1_url,
                user_id=self.good_user_2,
                user_pwd=self.good_user_2_pwd,
            )

            token = f_client.create_egeria_bearer_token(
                self.good_user_2, self.good_user_2_pwd
            )
            
            test_element_guid = "test-element-guid-123"
            
            start_time = time.perf_counter()
            response = f_client.get_attached_ratings(
                test_element_guid,
                start_from=0,
                page_size=50
            )
            duration = time.perf_counter() - start_time
            
            print(f"\n\tDuration was {duration} seconds")
            print(f"\n\tResponse was: {json.dumps(response, indent=2)}")
            
            assert True, "Ratings retrieved successfully"

        except PyegeriaInvalidParameterException as e:
            print_exception_table(e)
            assert False, "Invalid parameter exception"
        except PyegeriaAPIException as e:
            print_exception_table(e)
            assert False, "API exception"
        except PyegeriaUnauthorizedException as e:
            print_exception_table(e)
            assert False, "Unauthorized exception"
        except ValidationError as e:
            print_validation_error(e)
            assert False, "Validation error"
        except Exception as e:
            print_basic_exception(e)
            assert False, "Unexpected exception"
        finally:
            f_client.close_session()

    def test_remove_like_from_element(self):
        """Test removing a like from an element"""
        try:
            f_client = FeedbackManager(
                self.good_view_server_2,
                self.good_platform1_url,
                user_id=self.good_user_2,
                user_pwd=self.good_user_2_pwd,
            )

            token = f_client.create_egeria_bearer_token(
                self.good_user_2, self.good_user_2_pwd
            )
            
            test_element_guid = "test-element-guid-123"
            
            start_time = time.perf_counter()
            response = f_client.remove_like_from_element(
                test_element_guid,
                body={}
            )
            duration = time.perf_counter() - start_time
            
            print(f"\n\tDuration was {duration} seconds")
            print(f"\n\tResponse was: {json.dumps(response, indent=2)}")
            
            assert True, "Like removed successfully"

        except PyegeriaInvalidParameterException as e:
            print_exception_table(e)
            assert False, "Invalid parameter exception"
        except PyegeriaAPIException as e:
            print_exception_table(e)
            assert False, "API exception"
        except PyegeriaUnauthorizedException as e:
            print_exception_table(e)
            assert False, "Unauthorized exception"
        except ValidationError as e:
            print_validation_error(e)
            assert False, "Validation error"
        except Exception as e:
            print_basic_exception(e)
            assert False, "Unexpected exception"
        finally:
            f_client.close_session()

    def test_remove_rating_from_element(self):
        """Test removing a rating from an element"""
        try:
            f_client = FeedbackManager(
                self.good_view_server_2,
                self.good_platform1_url,
                user_id=self.good_user_2,
                user_pwd=self.good_user_2_pwd,
            )

            token = f_client.create_egeria_bearer_token(
                self.good_user_2, self.good_user_2_pwd
            )
            
            test_element_guid = "test-element-guid-123"
            
            start_time = time.perf_counter()
            response = f_client.remove_rating_from_element(
                test_element_guid,
                body={}
            )
            duration = time.perf_counter() - start_time
            
            print(f"\n\tDuration was {duration} seconds")
            print(f"\n\tResponse was: {json.dumps(response, indent=2)}")
            
            assert True, "Rating removed successfully"

        except PyegeriaInvalidParameterException as e:
            print_exception_table(e)
            assert False, "Invalid parameter exception"
        except PyegeriaAPIException as e:
            print_exception_table(e)
            assert False, "API exception"
        except PyegeriaUnauthorizedException as e:
            print_exception_table(e)
            assert False, "Unauthorized exception"
        except ValidationError as e:
            print_validation_error(e)
            assert False, "Validation error"
        except Exception as e:
            print_basic_exception(e)
            assert False, "Unexpected exception"
        finally:
            f_client.close_session()


if __name__ == "__main__":
    print("Running FeedbackManager unit tests...")


from __future__ import annotations

import json
import os
from typing import Dict, Any, Iterable

from .parse_compact_export import load_compact_json, expand_command


def _to_semicolon_str(value: Any) -> str:
    if value is None:
        return ""
    if isinstance(value, str):
        return value
    if isinstance(value, (list, tuple, set)):
        return "; ".join([str(v) for v in value if str(v).strip()])
    return str(value)


def _attr_to_legacy_entry(attr: Dict[str, Any]) -> Dict[str, Any]:
    # The legacy schema expects a one-item dict: { "Label": { ...metadata... } }
    label = attr.get("name") or attr.get("variable_name") or "Unknown"
    meta: Dict[str, Any] = dict(attr)
    # Remove wrapper key 'name' used only as the label
    meta.pop("name", None)
    return {label: meta}


def _command_to_legacy_spec(name: str, cdef: Dict[str, Any], expanded: Dict[str, Any]) -> Dict[str, Any]:
    # Copy known top-level fields if present in the compact command
    legacy: Dict[str, Any] = {}
    for key in (
        "display_name",
        "qn_prefix",
        "family",
        "description",
        "verb",
        "upsert",
        "attach",
        "ReferenceURL",
        "level",
        "find_method",
        "find_constraints",
        "extra_find",
        "extra_constraints",
        "Journal Entry",
    ):
        if key in cdef:
            legacy[key] = cdef[key]

    # Normalize alternate_names to the legacy semicolon string
    alt = cdef.get("alternate_names", "")
    legacy["alternate_names"] = _to_semicolon_str(alt)

    # Build legacy Attributes from expanded attribute definitions
    attrs_legacy = []
    for attr in expanded.get("all_attributes", []):
        attrs_legacy.append(_attr_to_legacy_entry(attr))
    legacy["Attributes"] = attrs_legacy

    return legacy




def load_compact_specs_from_dir(dir_path: str, families_allowlist: Iterable[str] | None = None) -> Dict[str, Dict[str, Any]]:
    """
    Load all compact command files from a directory and return a mapping suitable
    for merging into the legacy COMMAND_DEFINITIONS["Command Specifications"].

    Args:
        dir_path: Path to directory with compact JSON files.
        families_allowlist: If provided, include only commands whose 'family' is in this set.

    Returns:
        Dict[str, Dict]: { command_name: legacy_style_spec }
    """
    if families_allowlist:
        families = {f.strip() for f in families_allowlist if f and f.strip()}
    else:
        families = set()

    specs: Dict[str, Dict[str, Any]] = {}

    if not os.path.isdir(dir_path):
        return specs

    all_attr_defs = {}
    all_bundles = {}
    all_commands = {}

    for fname in os.listdir(dir_path):
        if not fname.endswith(".json"):
            continue
        fpath = os.path.join(dir_path, fname)
        try:
            data = load_compact_json(fpath)
        except Exception as e:
            from loguru import logger
            logger.warning(f"Failed to load compact JSON {fname}: {e}")
            continue

        all_attr_defs.update(data.get("attribute_definitions", {}))
        all_bundles.update(data.get("bundles", {}))
        all_commands.update(data.get("commands", {}))

    for cname, cdef in all_commands.items():
        # Family gating
        fam = cdef.get("family")
        if families and fam not in families:
            continue
        
        try:
            expanded = expand_command(cdef, all_bundles, all_attr_defs)
            specs[cname] = _command_to_legacy_spec(cname, cdef, expanded)
        except (KeyError, ValueError) as e:
            from loguru import logger
            logger.warning(f"Skipping command '{cname}': {e}")
            continue

    return specs



"""This lists all elements of an oopen metadata type"""

import argparse
import os
import sys
import time

from rich import box
from rich.console import Console
from rich.markdown import Markdown
from rich.prompt import Prompt
from rich.table import Table

from pyegeria import (
    EgeriaTech,
    PyegeriaException,
    print_basic_exception,
    SolutionArchitect,
    settings, load_app_config, pretty_print_config,
    config_logging,
    save_mermaid_html,
)

app_config = settings.Environment
config_path = os.path.join(app_config.pyegeria_config_directory, app_config.pyegeria_config_file)

EGERIA_USER = os.environ.get("EGERIA_USER", "erinoverview")
EGERIA_USER_PASSWORD = os.environ.get("EGERIA_USER_PASSWORD", "secret")
EGERIA_MERMAID_FOLDER = os.path.join(app_config.pyegeria_root, app_config.egeria_mermaid_folder)
conf = load_app_config(config_path)
# print(f"Loading config from {config_path} and mermaid folder is {EGERIA_MERMAID_FOLDER}")

config_logging()


def list_elements(
    om_type: str,
    server: str,
    url: str,
    username: str,
    password: str,
    jupyter: bool = app_config.egeria_jupyter,
    width: int = app_config.console_width,
):
    c_client = EgeriaTech(server, url, user_id=username, user_pwd=password)
    token = c_client.create_egeria_bearer_token()

    om_typedef = c_client.get_typedef_by_name(om_type)
    if type(om_typedef) is str:
        print(
            f"The type name '{om_type}' is not known to the Egeria platform at {url} - {server}"
        )
        sys.exit(1)
    elements = c_client.get_elements(om_type)

    def generate_table() -> Table:
        """Make a new table."""
        table = Table(
            caption=f"Metadata Elements for: {url} - {server} @ {time.asctime()}",
            style="bold bright_white on black",
            row_styles=["bold bright_white on black"],
            header_style="white on dark_blue",
            title_style="bold bright_white on black",
            caption_style="white on black",
            show_lines=True,
            box=box.ROUNDED,
            title=f"Elements for Open Metadata Type: '{om_type}' ",
            expand=True,
            # width=500
        )

        table.add_column("Qualified Name")
        table.add_column("Type")
        table.add_column("Created")
        table.add_column("Home Store")
        table.add_column("GUID", width=38, no_wrap=True)
        table.add_column("Properties")
        table.add_column("Other Classifications")

        if type(elements) is list:
            for element in elements:
                header = element["elementHeader"]
                el_q_name = element["properties"].get("qualifiedName", "---")
                el_type = header["type"]["typeName"]
                el_home = header["origin"]["homeMetadataCollectionName"]
                el_create_time = header["versions"]["createTime"][:-10]
                el_guid = header["guid"]
                el_class = header.get("otherClassifications", "---")

                el_props_md = ""
                for prop in element["properties"].keys():
                    el_props_md += f"* **{prop}**: {element['properties'][prop]}\n"
                el_props_out = Markdown(el_props_md)

                c_md = ""
                if type(el_class) is list:
                    for classification in el_class:
                        classification_name = classification.get(
                            "classificationName", "---"
                        )
                        c_md = f"* **{classification_name}**\n"
                        class_props = classification.get(
                            "classificationProperties", "---"
                        )
                        if type(class_props) is dict:
                            for prop in class_props.keys():
                                c_md += f"  * **{prop}**: {class_props[prop]}\n"
                c_md_out = Markdown(c_md)

                table.add_row(
                    el_q_name,
                    el_type,
                    el_create_time,
                    el_home,
                    el_guid,
                    el_props_out,
                    c_md_out,
                )

            return table
        else:
            print("No instances found")
            sys.exit(1)

    try:
        console = Console(width=width)
        with console.pager(styles=True):
            console.print(generate_table())

    except (
        PyegeriaException
    ) as e:
        print_basic_exception(e)
        print("\n\nPerhaps the type name isn't known")
    finally:
        c_client.close_session()


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--server", help="Name of the server to display status for")
    parser.add_argument("--url", help="URL Platform to connect to")
    parser.add_argument("--userid", help="User Id")
    parser.add_argument("--password", help="Password")

    args = parser.parse_args()

    server = args.server if args.server is not None else app_config.egeria_view_server
    url = args.url if args.url is not None else app_config.egeria_platform_url
    userid = args.userid if args.userid is not None else EGERIA_USER
    password = args.password if args.password is not None else EGERIA_USER_PASSWORD

    try:
        om_type = Prompt.ask(
            "Enter the Open Metadata Type to find elements of:", default="Referenceable"
        )
        list_elements(om_type, server, url, userid, password)
    except KeyboardInterrupt:
        pass


if __name__ == "__main__":
    main()
